import importlib as imp
import pandas as pd, numpy as np
import os,matplotlib,cbf,tifffile,re,glob,easygui
from subprocess import check_output
from matplotlib import rc
from crystals import Crystal
from scipy.spatial import ConvexHull, convex_hull_plot_2d
from utils import glob_colors as colors,handler3D as h3d
from utils import physicsConstants as cst
from scattering.structure_factor import structure_factor3D
from utils import displayStandards as dsp   ; imp.reload(dsp)
from . import rotating_crystal as rcc       #; imp.reload(rcc)
from . import postprocess as pp             #; imp.reload(pp)
from . import multi_3D as MS3D              #;imp.reload(MS3D)
from . import pymultislice                  #;imp.reload(pymultislice)
# from . import pets as pt                    #;imp.reload(pt)

cs = {1:colors.unicolor(0.75),3:(1,0,0),
      6:colors.unicolor(0.25),7:(0,0,1),8:(1,0,0),16:(1,1,0),14:(1,0,1),17:(0,1,1)}

################################################################################
#### Multislice related
################################################################################
def multi3D(name='./unknwon',filename='',load_opt=0,**kwargs):
    if filename :
        name=filename.replace('_3D.pkl','')
    else :
        filename=name+'_3D.pkl'

    if load_opt :
        if os.path.exists(filename):
            print(colors.green+'loading '+colors.yellow+filename+colors.black)
            return pymultislice.load(filename)
        else:
            print(colors.red+'file not found : '+colors.yellow+filename+colors.black)
    return MS3D.Multi3D(name=name,**kwargs)

#
# def sweep_var(datpath,param,vals,df=None,ssh='',tail='',pargs=True,do_prev=0,
#     **kwargs):
#     from . import multislice as mupy            ; imp.reload(mupy)
#     '''
#     runs a set of similar simulations with one varying parameter
#     - datpath    : path to the simulation folder
#     - param,vals : the parameters and values to sweep
#     - df :
#         - int - create and save the new dataframe if 1
#         - pd.Dataframe - to update(since parsed as a reference)
#     - pargs         : bool - pass param and values to Multislice  if True
#     - do_prev       : Used for iterative fourier transform
#     - kwargs : see help(Multislice)
#     '''
#     do_df,save = isinstance(df,pd.core.frame.DataFrame),0
#     if isinstance(df,int):
#         if df : df,do_df,save = pd.DataFrame(columns=[param,'host','state']+pp.info_cols),1,1
#     nvals,prev = len(vals),None
#     for i,val in zip(range(nvals),vals):
#         if pargs:kwargs[param]=val
#         if do_prev and i: prev = multi.outf['image']
#         multi=mupy.Multislice(datpath,prev=prev,
#             ssh=ssh,tail=tail+param+str(i).zfill(int(np.ceil(nvals/10))),
#             **kwargs)
#         if do_df:
#             df.loc[multi.outf['obj']] = [np.nan]*len(df.columns)
#             df.loc[multi.outf['obj']][[param,'host','state']] = [val,ssh,'start']
#     if save :
#         df.to_pickle(datpath+'df.pkl')
#         print(colors.green+'Dataframe saved : '+colors.yellow+datpath+'df.pkl'+colors.black)
#         return df

################################################################################
#### Reciprocal space related
################################################################################
def get_reciprocal(abc):
    a1,a2,a3 = abc
    b1 = np.cross(a2,a3)/(a1.dot(np.cross(a2,a3)))
    b2 = np.cross(a3,a1)/(a2.dot(np.cross(a3,a1)))
    b3 = np.cross(a1,a2)/(a3.dot(np.cross(a1,a2)))
    abc_star = np.vstack([b1,b2,b3])#*(2*np.pi)
    return abc_star

def get_lattice(lat_vec,Nmax=5):
    '''Get a lattice up to Nmax points :
    - lat_vec : reciprocal lattice vectors
    - Nmax : max order '''
    a1,b1,c1 = lat_vec
    N = np.arange(-Nmax,Nmax+1)
    h,k,l = np.meshgrid(N,N,N)
    h,k,l = h.flatten(),k.flatten(),l.flatten()
    qx = h*a1[0]+k*b1[0]+l*c1[0]
    qy = h*a1[1]+k*b1[1]+l*c1[1]
    qz = h*a1[2]+k*b1[2]+l*c1[2]
    return (h,k,l),(qx,qy,qz)
def get_lattice2D(lat_vec,Nmax=5):
    '''Get a lattice up to Nmax points :
    - lat_vec : reciprocal lattice vectors
    - Nmax : max order '''
    a1,b1 = lat_vec
    N = np.arange(-Nmax,Nmax+1)
    h,k = np.meshgrid(N,N)
    h,k = h.flatten(),k.flatten()
    qx = h*a1[0]+k*b1[0]
    qz = h*a1[1]+k*b1[1]
    return (h,k),(qx,qz)

def get_ewald(K,nts=100,nps=200):
    ''' Get ewald sphere coordinates
    - K : reciprocal space beam vector
    '''
    Kx,Ky,Kz = K

    theta,phi = np.linspace(0,np.pi,nts),np.linspace(0,2*np.pi,nps)
    x = Kx+self.K0*np.outer(np.sin(theta),np.cos(phi))
    y = Ky+self.K0*np.outer(np.sin(theta),np.sin(phi))
    z = Kz+self.K0*np.outer(np.cos(theta),np.ones(phi.shape))
    return x,y,z

def get_excitation_errors(K,lat_vec,Nmax=5,Smax=0.02):
    ''' get excitation errors for lattice lat_vec and beam K
    - K : reciprocal space beam vector
    - Nmax : max order of reflections(resolution)
    - Smax : maximum excitation error to be included
    '''
    K0 = np.linalg.norm(K)
    Kx,Ky,Kz = K

    (h,k,l),(qx,qy,qz) = get_lattice(lat_vec,Nmax)

    Sw = np.abs(np.sqrt((Kx-qx)**2+(Ky-qy)**2+(Kz-qz)**2) - K0)
    if Smax:
        idx = Sw<Smax
        h,k,l = np.array([h[idx],k[idx],l[idx]],dtype=int)
        qx,qy,qz,Sw = qx[idx],qy[idx],qz[idx],Sw[idx]
    d = dict(zip(['h','k','l','qx','qy','qz','Sw'],[h,k,l,qx,qy,qz,Sw]))
    return pd.DataFrame.from_dict(d)

def get_structure_factor(cif_file,dfout=0,**sf_args):
    '''computes structure factor 3D
    - sf_args : see (structure_factor3D)
    returns :
    - (qx,qy,qz),Fhkl
    '''
    crys = import_crys(cif_file)
    pattern = np.array([np.hstack([a.coords_fractional,a.atomic_number]) for a in crys.atoms] )
    lat_vec = np.array(crys.reciprocal_vectors)
    (h,k,l),Fhkl = structure_factor3D(pattern, lat_vec, **sf_args)
    qx = h/crys.lattice_parameters[0]
    qy = k/crys.lattice_parameters[1]
    qz = l/crys.lattice_parameters[2]
    if dfout:
        data = [h.flatten(),k.flatten(),l.flatten(),qx.flatten(),qy.flatten(),qz.flatten(),Fhkl.flatten()]
        d = dict(zip(['h','k','l','qx','qy','qz','Fhkl'],data))
        return pd.DataFrame.from_dict(d)
    else:
        return (qx,qy,qz),Fhkl

def get_excited_beams(cif_file,K,lat_vec,Nmax,Smax):
    df_Sw   = get_excitation_errors(K,lat_vec,Nmax,Smax)
    df_Fhkl = get_structure_factor(cif_file,dfout=1,hklMax=Nmax+1)

    hkl0 = df_Sw[['h','k','l']].values
    hklF = df_Fhkl[['h','k','l']].values
    ridx = []
    for r in hkl0:
        idx = np.where(np.linalg.norm(r-hklF,axis=1)==0)[0]
        if idx.size:
            ridx+=[idx[0]]

    Fhkl_0 = df_Fhkl.iloc[ridx]['Fhkl'].values #;print(df_Fhkl.iloc[ridx]);#print(ridx);print(df_Fhkl.shape)
    df_Sw['Fhkl'] = Fhkl_0
    return df_Sw

def get_kinematic_intensities(cif_file,K,thick,Nmax=5,Smax=None):
    crys    = import_crys(cif_file)
    lat_vec = np.array(crys.reciprocal_vectors)/(2*np.pi)
    df_Sw   = get_excited_beams(cif_file,K,lat_vec,Nmax,Smax)

    #kinematic intensities
    K0  = np.linalg.norm(K)
    sig = cst.keV2sigma(cst.lam2keV(1/K0))      #;print(sig)
    Sw  = df_Sw.Sw.values
    gs = (sig*np.pi*thick*Sw)*np.sinc(thick*Sw)

    Fhkl_0 = df_Sw['Fhkl'].values
    I  = (np.abs(Fhkl_0)*gs)**2
    df_Sw['I'] = I
    return df_Sw

def project_beams(K,qxyz,e0=[1,0,0],v=0):
    e3 = K/np.linalg.norm(K)
    e0 = e0/np.linalg.norm(e0)
    e2 = np.cross(e3,e0)
    e1 = np.cross(e2,e3)

    px,py = qxyz.dot(e1),qxyz.dot(e2)
    if v:
        return px,py,e0.dot(e1)
    else:
        return px,py

def project_beams2D(K,qxy):
    u = K/np.linalg.norm(K)
    v = np.array([u[1],-u[0]])
    px = qxy.dot(v)
    return px

################################################################################
#### coordinate file generation from cif files
################################################################################
import wallpp.lattice as lat        ;imp.reload(lat)
import wallpp.plane_group as pg     ;imp.reload(pg)
# from wallpp import wallpaper as pp  ;imp.reload(pp)
def import_wallpp(file,**kwargs):
    ''' load file into a wallpp
    - file  : str -
        - .py python file containing pptype,ax,bz,angle,pattern
        - .txt
    '''
    if isinstance(file,dict):
        pp_args=file
    else:
        ext = file.split('.')[-1]
        with open(file,'r') as f:lines=f.readlines()
        if ext =='py':
            d={}
            exec(''.join(lines),d) #pptype,ax,bz,angle,pattern defined in the lines
            pptype,ax,bz,angle = [d[k] for k in ['pptype','ax','bz','angle']]
            pattern=np.array(d['pattern'])
        elif ext=='txt':
            lines = [l.strip() for l in lines]
            pptype = lines[0]
            ax,bz,angle = np.array(lines[1:4],dtype=float)
            pattern = np.array([np.array(l.split(' '),dtype=float) for l in lines[4:]])
        pp_args=dict(zip(
            ['pp_type','a','b','angle','pattern'],
            [pptype,ax,bz,angle,pattern]))
    wallpp = pg.Wallpaper(**pp_args,**kwargs)
    # wallpp.reciprocal_vectors = np.array(wallpp.get_reciprocal_lattice_2D())
    # wallpp.lattice_vectors = np.array(wallpp.lattice_vec)
    # wallpp.pattern=pattern #np.hstack([wallpp.Xa,wallpp.fa[:,None]])
    # a1,a2 = wallpp.lattice_vectors
    # wallpp.area=abs(np.cross(a1,a2))
    return wallpp

def gen_xyz2(file,xyz,lat_params,n=[0,0,1],theta=0, pad=0,fmt='%.4f',opts=''):
    if 'v' in opts:print('...import file...')
    crys = import_crys(file)
    n_u = n
    pattern = np.array([[a.atomic_number]+list(a.coords_cartesian)+[a.occupancy,1.0] for a in crys.atoms])
    lat_vec = np.array(crys.lattice_vectors)
    coords = pattern[:,1:4]
    Za,occ,bfact = pattern[:,[0,4,5]].T
    ax,by,cz = lat_params

    #### replicate
    if 'v' in opts:print('...finding unit cells index...')
    l,m,n = find_xyz(lat_vec,lat_params,n_u,theta,plot='p' in opts,v='v' in opts)
    a1,a2,a3 = np.array(lat_vec)
    coords = np.vstack([coords + i*a1+j*a2+k*a3 for i,j,k in zip(l,m,n)])
    Za     = np.tile(Za   ,[n.size])
    occ    = np.tile(occ  ,[n.size])
    bfact  = np.tile(bfact,[n.size])

    ###orient
    coords = rcc.orient_crystal(coords,n_u=n_u,theta=theta)#;print(coords.shape)
    x,y,z = coords.T
    idx = (x>=0) & (y>=0) & (z>=0) & (x<=ax) & (y<=by) & (z<=cz)
    # coords=coords[idx]
    # print(coords.shape)
    #### apply padding
    if isinstance(pad,int) or isinstance(pad,float):pad=[pad]*2
    if sum(pad)>0:
        coords[:,0] += ax*pad[0]
        coords[:,1] += by*pad[1]
        ax *= 1+2*pad[0]
        by *= 1+2*pad[1]

    pattern = np.hstack([Za[idx,None],coords[idx],occ[idx,None],bfact[idx,None]])
    #### save
    if 'v' in opts:print('...saving to file ...')
    dir=''.join(np.array(n,dtype=str))
    header = 'one unit cell of %s at %s, %.1f degree\n' %(dsp.basename(file),str(n_u),theta)
    header+= ' '.join([fmt]*3) %(ax,by,cz)
    np.savetxt(xyz,pattern,footer='-1',header=header,fmt='%d '+' '.join([fmt]*5),comments='')
    print(colors.green+"coords file saved : \n"+colors.yellow+xyz+colors.black)
    npy_file = xyz.replace('.xyz','.npy')
    np.save(npy_file,[(ax,by,cz),pattern])
    print(colors.green+'binary coords file saved :'+colors.yellow+npy_file+colors.black)
    # if 'v' in opts:
    print('number of coordinates = %d' %pattern.shape[0])
    print('number of unit cells  = %d' %l.shape[0])

def find_xyz(lat_vec,lat_params,n_u,theta,plot=0,v=0):
    ax,by,cz = lat_params
    ai = np.linalg.norm(lat_vec,axis=1)#;print(a_max)
    a_min,a_max = min(ai),max(ai)
    # ra1,ra2,ra3 = rot_vec
    #### brute force unit cells generation
    N = np.ceil(1.5*max(lat_params)/a_min)
    u1 = np.arange(-N,N+1)
    l,m,n = np.meshgrid(u1,u1,u1)
    lmn = np.vstack([l.flatten(),m.flatten(),n.flatten()]).T
    if v:print('...\torienting full mesh N=%d...' %N)
    #### l*a1+m*a2+n*a3
    rot_vec = rcc.orient_crystal(lat_vec,n_u=n_u,T=True,theta=theta)
    xyz  = lmn.dot(rot_vec)
    x,y,z = xyz.T
    if v:print('...\tkeeping valid unit cells...')
    idx = (x>=-a_max) & (y>=-a_max) & (z>=-a_max) & (x<=ax+a_max) & (y<=by+a_max) & (z<=cz+a_max)
    lmn = lmn[idx]

    if plot:
        print('N',N)
        print('rot_vec',rot_vec)
        l,m,n = lmn.T
        x,y,z = xyz[idx].T
        print('lminmax',l.min(),l.max(),m.min(),m.max(),n.min(),n.max())

        #### generate super cell corners
        l0,m0,n0 = np.meshgrid([0,1],[0,1],[0,1])
        ax_,by_,cz_ = np.diag(lat_params)
        sc = np.vstack([ax_*i+by_*j+cz_*k for i,j,k in zip(l0.flatten(),m0.flatten(),n0.flatten())])
        x0,y0,z0 = sc.T
        #### get actual corners
        lat_p2 = np.linalg.norm(rot_vec,axis=1)**2
        nlm1 = np.array([np.round(rot_vec.dot(v)/lat_p2) for v in sc])
        print('corners' ,nlm1)
        x1,y1,z1 = nlm1.dot(rot_vec).T

        # scat = ()
        # scat+=([x0,y0,z0,50,'b','o'],)
        # scat+=([x0,y0,z0,50,'g','d'],)
        # scat+=([x,y,z,20,'r','s'],)
        # dsp.stddisp(rc='3d',scat=scat)
        fig,(ax0,ax1) = dsp.create_fig(rc='12')
        scat0 = [x,y,30,'b']
        scat1 = [x,z,30,'b']
        pps0 = [dsp.Rectangle((0,0),ax,by,linewidth=2,edgecolor='b',alpha=0.1)]
        pps1 = [dsp.Rectangle((0,0),ax,cz,linewidth=2,edgecolor='b',alpha=0.1)]
        dsp.stddisp(ax=ax0,scat=scat0,patches=pps0,opt='',axPos=0)
        dsp.stddisp(ax=ax1,scat=scat1,patches=pps1,opt='',axPos=0)
        fig.show()
    return lmn.T

def gen_xyz(file,n=[0,0,1],rep=[1,1,1],pad=0,xyz='',theta=0,**kwargs):
    ''' convert cif file into autoslic .xyz input file
    - file : cif_file
    - rep : super cell repeat
    - n : reorient of the z axis into n
    - pad : amount of padding on each side (in unit of super cell size)
    '''
    if isinstance(file,str):
        tail = ''.join(np.array(n,dtype=str)) + '.xyz'
        if file.split('.')[-1]=='cif':
            crys = Crystal.from_cif(file)
            if not xyz : xyz = file.replace('.cif',tail)
        elif sum(np.array(list(Crystal.builtins))==file):
            crys = Crystal.from_database(file)
            if not xyz : xyz = file+tail
        lat_vec    = np.array(crys.lattice_vectors)
        lat_params = crys.lattice_parameters[:3]
        pattern    = np.array([[a.atomic_number]+list(a.coords_cartesian)+[a.occupancy,1.0] for a in crys.atoms])
    else:
        if not xyz: raise Exception('xyz filename required')
        pattern = file
    pattern,lat = make_xyz(xyz,pattern,lat_vec,lat_params,n=n,pad=pad,rep=rep,theta=theta,**kwargs)
    npy_file = xyz.replace('.xyz','.npy')
    np.save(npy_file,[lat,pattern])
    print(colors.green+'binary coords file saved :'+colors.yellow+npy_file+colors.black)

def import_cif(file,xyz='',n=[0,0,1],rep=[1,1,1],pad=0,dopt='s',lfact=1.0,tail=''):
    """ convert cif file into autoslic .xyz input file

    Parameters
    ----------
    file
        cif_file
    rep
        super cell repeat
    n
        reorient of the z axis into n
    pad
        amount of padding on each side (in unit of super cell size)
    """
    crys       = Crystal.from_cif(file)
    lat_vec    = np.array(crys.lattice_vectors)
    lat_params = crys.lattice_parameters[:3]
    pattern    = np.array([[a.atomic_number]+list(lfact*a.coords_cartesian)+[a.occupancy,1.0] for a in crys.atoms])

    if xyz:make_xyz(xyz,pattern,lat_vec,lat_params,n=n,pad=pad,rep=rep,fmt='%.4f',dopt=dopt)
    pattern[:,1:4] = rcc.orient_crystal(pattern[:,1:4],n_u=n) #,lat_params #pattern,crys # file
    return pattern


def make_xyz(name,pattern,lat_vec,lat_params,n=[0,0,1],theta=0,rep=[1,1,1],pad=0,fmt='%.4f',dopt='s'):
    '''Creates the.xyz file from a given compound and orientation
    - name       : Full path to the file to save
    - pattern    : Nx6 ndarray - Z,x,y,z,occ,wobble format
    - lat_vec    : 3x3 ndarray - lattice vectors [a1,a2,a3]
    - lat_params : ax,by,cz
    - pad  : amount of padding on each side (in unit of super cell size)
    - n    : beam direction axis
    - dopt : p(print file),s(save)
    '''
    compound = dsp.basename(name)
    Za,occ,bfact = pattern[:,[0,4,5]].T
    coords = pattern[:,1:4]
    Nx,Ny,Nz = rep
    ax0,by0,cz0 = lat_params
    ax,by,cz = lat_params
    #replicate
    if sum(rep)>3 :
        ni,nj,nk = np.meshgrid(range(Nx),range(Ny),range(Nz))
        ni,nj,nk = ni.flatten(),nj.flatten(),nk.flatten()
        a1,a2,a3 = np.array(lat_vec)
        coords = np.vstack([coords + i*a1+j*a2+k*a3 for i,j,k in zip(ni,nj,nk)])
        Za     = np.tile(Za   ,[ni.size])
        occ    = np.tile(occ  ,[ni.size])
        bfact  = np.tile(bfact,[ni.size])
        ax = Nx*ax0
        by = Ny*by0
        cz = Nz*cz0

    #orient
    coords = rcc.orient_crystal(coords,n_u=n)
    if theta :
        t = np.deg2rad(theta)
        st,ct = np.sin(t),np.cos(t)
        coords = np.array([[ct,st,0],[-st,ct,0],[0,0,1]]).dot(coords.T).T
    #apply padding
    if isinstance(pad,int) or isinstance(pad,float):pad=[pad]*2
    if sum(pad)>0:
        coords[:,0] += Nx*ax0*pad[0]
        coords[:,1] += Ny*by0*pad[1]
        # coords[:,2] += Nz*cz0
        ax *= 1+2*pad[0]
        by *= 1+2*pad[1]
        # cz *= 1+2*pad[2]
    pattern = np.hstack([Za[:,None],coords,occ[:,None],bfact[:,None]])
    # ax,by,cz = lat_params
    lat_params = (ax,by,cz)
    #write to file
    if 's' in dopt :
        dir=''.join(np.array(n,dtype=str))
        header = 'one unit cell of %s\n' %(compound)
        header+= ' '.join([fmt]*3) %(ax,by,cz)
        np.savetxt(name,pattern,footer='-1',header=header,fmt='%d '+' '.join([fmt]*5),comments='')
        print(colors.green+"coords file saved : \n"+colors.yellow+name+colors.black)
        if 'p' in dopt :
            with open(name,'r') as f : print(''.join(f.readlines()))
    return pattern,[ax,by,cz]

def make_mulslice_datfile(dat_file,cif_file):
    ''' create a data file used by atompot(mulslice) from a cif file
    - dat_file : name of file to save
    - cif_file : the file to import
    '''
    crys = Crystal.from_cif(cif_file)
    pattern = np.array([np.hstack([a.coords_cartesian,a.atomic_number]) for a in crys.atoms] )
    deck = ' '.join(np.array(crys.lattice_parameters[:3],dtype=str))+'\n'
    deck+='0\n'
    Za = np.array(np.unique(pattern[:,-1]),dtype=int)
    for Z in Za:
        deck+=str(Z)+'\n'#; print()
        xyz = np.array(pattern[pattern[:,-1]==1][:,:3],dtype=str)
        xyz = '\n'.join([' '.join(l) for l in xyz]) #; print(xyz)
        deck+=xyz+'\n'
    deck+='\n\nComment here'
    with open(dat_file,'w') as f : f.write(deck)

def get_unit_cell(cell_mesh,n):
    alpha,lw,c = 0.1,2,'c'
    x,y,z = cell_mesh
    planes_coords,ncells = [], x.shape[0]
    for i in range(ncells):
        planes_coords+=[[x[i,:,:],y[i,:,:],z[i,:,:]],
                        [x[:,i,:],y[:,i,:],z[:,i,:]],
                        [x[:,:,i],y[:,:,i],z[:,:,i]]]
    surfs = []
    for plane_coords in planes_coords :
        x,y,z  = plane_coords
        p_shape = x.shape
        coords = np.array([x.flatten(),y.flatten(),z.flatten()]);#print(coords.shape)
        x,y,z  = rcc.orient_crystal(coords,n_u=n,T=False);
        x,y,z  = np.reshape(x,p_shape),np.reshape(y,p_shape),np.reshape(z,p_shape)
        surfs += [[x,y,z,c,alpha,lw,c]]
    return surfs

def get_arrow_3d(n,x0,rc=0.1,h=0.2):
    '''for trihedron'''
    nu,nv = 15,2; #print(u)
    shape = (nu,nv)
    u = np.linspace(0,2*np.pi,nu)
    v = np.linspace(0,np.pi/2,nv)
    x = np.outer(np.cos(u),np.sin(v))*h/2
    y = np.outer(np.sin(u),np.sin(v))*h/2
    z = np.outer(np.ones(u.shape),np.cos(v))*h
    coords = np.array([x.flatten(),y.flatten(),z.flatten()]);#print(coords.shape)
    #uvec = np.array(n);print(n)
    x,y,z = rcc.orient_crystal(coords,n,[0,0,1],T=False)
    x,y,z = np.reshape(x,shape),np.reshape(y,shape),np.reshape(z,shape)

    O = np.array([0,0,0]);#print(x0)#,x0 + np.array([O,n]))
    xl,yl,zl = (x0 + np.array([O,n])).T ;#print(xl,yl,zl)
    return [x+x0[0]+n[0],y+x0[1]+n[1],z+x0[2]+n[2]],[xl,yl,zl]

def get_vec(n,crys,bopt) :
    if isinstance(n,int) : n = crys.lattice_vectors[n]
    if bopt : n = np.array(n).dot(np.array(crys.lattice_vectors))
    return n


################################################################################
#### display coordinate file
################################################################################
def show_cell(file,n=[0,0,1],bopt=1,x0=None,rep=[1,1,1],h3D=1,
    xylims=None,axPos=[],**kwargs):
    '''Show unit cell and coords from cif file '''
    crys = Crystal.from_cif(file)
    n    = get_vec(n,crys,bopt)

    #unit cells,lattice vectors,atom coordinates
    cell_mesh = crys.mesh(range(rep[0]+1),range(rep[1]+1),range(rep[2]+1))
    surfs     = get_unit_cell(cell_mesh,n)
    uvw       = rcc.orient_crystal(np.array(crys.lattice_vectors),n_u=n,T=True)
    pattern   = import_cif(file,n=n,rep=rep)
    E,X,Y,Z   = pattern[:,:4].T
    scat      = [X,Y,Z, [cs[int(e)] for e in E]]

    #display options
    if x0==None : x0 = -1
    if isinstance(x0,float) or isinstance(x0,int) : x0=np.array([x0]*3)
    if not xylims:
        c1,c2,c3 = (X.min()+X.max())/2,(Y.min()+Y.max())/2,(Z.min()+Z.max())/2
        w = 0.75*max(X.max()-X.min(),Y.max()-Y.min(),Z.max()-Z.min())
        xylims=[c1-w/2,c1+w/2,c2-w/2,c2+w/2,c3-w/2,c3+w/2]
        # print(xylims)
    if not axPos:axPos=[0,0,1,0.95]
    fig,ax=dsp.stddisp(scat=scat,ms=100,surfs=surfs,rc='3d',std=0)
    show_trihedron(ax,uvw=uvw,x0=[0,0,0],cs=['r','g','b'],labs=['$a$','$b$','$c$'],lw=2,rc=0.1,h=0.2)
    show_trihedron(ax,x0=x0,labs=['$x$','$y$','$z$'],lw=2,rc=0.1,h=0.2)
    dsp.stddisp(ax=ax,xylims=xylims,axPos=axPos,
        pOpt='Xpt',**kwargs)

    if h3D:hdl = h3d.handler_3d(fig,persp=False)

def show_grid(file,opts='',popts='pv',figs='21',**kwargs):
    '''
    file : path to .xyz file
    opt : str format 'x1x2' - 'xy','xz','yz','zx',...
    '''
    if 'v' in popts:print('...loading file...')
    npy_file = file.replace('.xyz','.npy')
    if os.path.exists(npy_file):
        lat_params,pattern = np.load(npy_file,allow_pickle=True)
    else:
        with open(file,'r') as f:
            l=list(map(lambda s:s.strip().split(' '),f.readlines()))
            lat_params = np.array(l[1],dtype=float)
            pattern   = np.array(l[2:-1],dtype=float)
            np.save(npy_file,[lat_params,pattern])
            print(colors.green+'binary coords file saved :'+colors.yellow+npy_file+colors.black)

    if isinstance(opts,list):
        fig,axs = dsp.create_fig(figsize=figs,rc=[1,len(opts)])
        for axi,opts_i in zip(axs,opts):
            plot_grid(pattern,lat_params,opts_i,popts=popts,ax=axi,setPos=0,opt='',**kwargs)
        fig.show()
    else:
        plot_grid(pattern,lat_params,opts,**kwargs)

def plot_grid(pattern,lat_params,opts,popts='pv',xylims=[],**kwargs):
    #a1,a2,a3,alpha,beta,gamma=crys.lattice_parameters
    xij = {'x':1,'y':2,'z':3}
    vopt = 'v' in popts
    if opts:
        x1,x2 = opts
        i,j = xij[x1],xij[x2]
        Z = np.array(pattern[:,0],dtype=int)
        # print('...assigning colors...')
        C = np.array([cs[E] for E in Z])
        pps = [dsp.Rectangle((0,0),lat_params[i-1],lat_params[j-1],linewidth=2,edgecolor='b',alpha=0.1)]
        scat=[]
        if 'p' in popts:
            scat = [pattern[:,i],pattern[:,j],C]
        if 'h' in popts :
            if vopt:print('...finding convex hull...')
            hull = ConvexHull(pattern[:,[i,j]])#,incremental=True)
            idx = hull.vertices #;print(idx)
            points = np.array([pattern[idx,i],pattern[idx,j]]).T
            pps+=[dsp.matplotlib.patches.Polygon(points,linewidth=2,facecolor='r',edgecolor='r',alpha=0.2)]
        if vopt:print('...plotting...')
        if not xylims : xylims = [0,lat_params[i-1],0,lat_params[j-1]]
        return dsp.stddisp(labs=['$%s$'%x1,'$%s$'%x2],patches=pps,scat=scat,
            xylims=xylims,**kwargs)

def import_xyz(xyz_file):
    '''import .xyz file
    - xyz_file : file to import
    returns :
    - pattern    : [Za,coords,occ,bfactor]
    - lat_params : [ax,by,cz] for the supercell
    '''
    with open(xyz_file,'r') as f:l=list(map(lambda s:s.strip().split(' '),f.readlines()))
    lat_params = np.array(l[1],dtype=float)
    pattern = np.array(l[2:-1],dtype=float)
    return pattern,lat_params

def show_grid3(xyz_file,ms=1,**kwargs):
    '''display an .xyz file content in a 2D plane
    xyz_file : .xyz file
    '''
    pattern,lat_params = import_xyz(xyz_file)
    Z = np.array(pattern[:,0],dtype=int)
    C = [cs[E] for E in Z]

    scat = [pattern[:,1],pattern[:,2],pattern[:,3],ms,C]
    return dsp.stddisp(scat=scat,labs=['$x$','$y$','$z$'],rc='3d',**kwargs)

def show_trihedron(ax,uvw=None,x0=[0,0,0],cs=None,clabs=None,lw=2,rc=0.1,h=0.2,**kwargs):
    '''
    x0 : position of trihedron
    uvw : Nx3 ndarray
    ll,rc,h : length,radius and height af arrow/cones
    cs,labs : colors and labels of cones (default black and None)
    '''
    if not isinstance(uvw,np.ndarray) : uvw = np.identity(3)
    txts,x0=[],np.array(x0)
    if not cs : cs=['k']*uvw.shape[0]
    if clabs :
        xtxt = x0 + 1.1*uvw
        for i in range(len(clabs)):
            xt0,yt0,zt0 = xtxt[i,:]
            txts += [[xt0,yt0,zt0,clabs[i],cs[i][0]]]
    plots,surfs = [],[]
    for u,cu in zip(uvw,cs) :
        (x,y,z),(xl,yl,zl) = get_arrow_3d(u,x0,rc=0.1,h=0.2)
        plots += [[xl,yl,zl,cu]]
        surfs += [[x,y,z,cu[0],None,lw,cu[0]]]
    dsp.stddisp(ax=ax,texts=txts,plots=plots,surfs=surfs,lw=lw,**kwargs)

def show_ewald_sphere(lat_params,lam=0.025,tmax=7,T=0.2,nx=20,ny=10,**kwargs):
    '''lam(wavelength Angstrum), tmax(angle degrees), T(Thickness mum)'''
    a1,a2,a3 = lat_params
    b1,b2,b3 = 1/a1,1/a2,1/a3
    tmax, K=np.deg2rad(tmax), 1/lam
    dqT = 10/(T*1e4)*np.array([-1,1])
    #reciprocal lattice
    h,k = np.meshgrid(b1*np.arange(-nx,nx+1),b2*np.arange(-1,ny))
    #sphere
    t = 3*np.pi/2+np.linspace(-tmax,tmax,100)
    plts = [[K*np.cos(t),K*np.sin(t)+K,'r',r'$\lambda=%.2f\AA$' %lam]]
    #rel rods
    plts += [[ [h,h],k+dqT ,'b-',''] for h,k in zip(h.flatten(),k.flatten())]

    #display
    scat=[h,k,15,'b']
    fig,ax = dsp.stddisp(plts,scat=scat,labs=['$q_x$','$q_y$'],
        lw=3,#,xylims=[-nx*b1,nx*b1,-b2,ny*b2],xyTickLabs=[[],[]],
        **kwargs)


def show_Ewald2D(K,lat,relrod=0,**kwargs):
    ''' Displays the Ewald sphere and the reciprocal lattice
    - K : beam
    - lat : reciprocal lattice qx,qz
    '''
    #lattice
    qx,qz = lat
    qxmax = 1.1*max(abs(qx.max()),abs(qz.max()))
    scat  = [qx,qz,15,'k']

    # Ewald circle
    Kx,Kz = K
    k0 = np.linalg.norm(K)
    v = np.array([-Kz,Kx])/k0
    A1,A2 = qxmax*v,-qxmax*v

    t = np.linspace(0,2*np.pi,1000)
    plts = [[Kx + k0*np.cos(t),Kz +k0*np.sin(t),'m','Ewald circle']]
    plts += [[ [Kx,0],[Kz,0],'r','$K$']]
    plts += [[ [A1[0],A2[0]],[A1[1],A2[1]]  ,'b','detector']]

    # if relrod:
    #     H    = Nz*bz
    #     zeta = np.linspace(0,0.25/bz,100)
    #     Fz   = lambda i : 1/(1.1*self.ax)*np.sinc(zeta*H)**2+i/self.ax
    #     plts +=[[Fz(i),zeta,'b--',''] for i in range(nh)]
    return dsp.stddisp(plts,scat=scat,
        lw=2,labs=[r'$q_x(\AA^{-1})$',r'$q_z(\AA^{-1})$'],
        xylims=qxmax,
        **kwargs)#lw=2,xylims=[0,3,0,3])#,xyTicks=[1/ax1,1/bz1])










#
# ################################################################################
# #### Class viewers
# ################################################################################
# class Image_viewer:
#     '''a copy of dials.image_viewer'''
#     def __init__(self,file,sym=1,pad=None):
#         basename      = file.split('_')
#         self.im       = int(basename[-1].replace('.cbf',''))
#         self.basename = '_'.join(basename[:-1])
#         file_ids      = [int(f.split('_')[-1].replace('.cbf','')) for f in glob.glob(self.basename+'*.cbf')]
#         if pad :
#             self.pad=pad
#         else:
#             self.pad=int(np.log10(max(file_ids)))+1       #;print(self.pad)
#
#         ls = np.array(check_output("head -n25 %s" %file,shell=True).decode().split('\n'))
#         px_str      = ls[['Pixel_size' in l for l in ls]][0]
#         D_str       = ls[['Detector_distance' in l for l in ls]][0] #;print(D_str)
#         lam_str     = ls[["Wavelength" in l for l in ls]][0]
#
#         self.pxy    = np.array(re.findall("\d+e-\d+", px_str),dtype=float)
#         self.D      = 1 #float(re.findall("\d+.\d+",D_str)[0])
#         self.lam    = float(re.findall("\d+.\d+",lam_str)[0])
#
#         shape = np.array(cbf.read(file).data.shape)
#         self.image = cbf.read(file).data
#
#         if sym:
#             Nx,Ny = np.array(shape/2,dtype=int)
#             self.Nx,self.Ny = Nx,Ny
#             self.pX,self.pY = np.meshgrid(np.arange(-Nx,Nx+1),np.arange(-Ny,Ny+1))
#         else:
#             Nx,Ny = shape
#             # self.pX,self.pY = np.meshgrid(np.arange(Nx),np.arange(Ny))
#             self.pX,self.pY = np.meshgrid(np.arange(Nx),np.flipud(np.arange(Ny)))
#         self.qx,self.qy = self.px2s(self.pX,self.pY)
#
#     def s2px(self,xh):
#         '''xh : recorded coordinates on image'''
#         px,py = self.pxy
#         pX = self.lam*self.D/px*xh[:,0] + self.Nx
#         pY = -self.lam*self.D/py*xh[:,1] + self.Ny
#         return pX,pY
#
#     def px2s(self,pX,pY):
#         px,py = self.pxy
#         qx,qy = px/self.D/self.lam*pX, py/self.D/self.lam*pY
#         return qx,qy
#
#     def show_image(self,im=None,lab='q',stack=1,**kwargs):
#         rc('text', usetex=False)
#         if not im:im = self.im
#         file = "%s_%s.cbf" %(self.basename,str(im).zfill(self.pad))   #;print(filename)
#         # with open(file,'wb') as f:print(file)
#         content = cbf.read(file)
#         image = content.data
#         if stack>1:
#             for i in range(1,stack):
#                 filename = "%s_%s.cbf" %(self.basename,str(im+i).zfill(self.pad))   #;print(filename)
#                 image+=cbf.read(filename).data
#             if 'caxis' in list(kwargs.keys()):
#                 kwargs['caxis'] = list(np.array(kwargs['caxis'])*stack) #; print(kwargs['caxis'])
#         if 'q' in lab:
#             labs = ['$q_x(A^{-1})$','$q_y(A^{-1})$']
#             X,Y = self.qx,-self.qy
#         elif 'p' in lab:
#             labs = ['','']#['$p_x$','$p_y$']
#             X,Y = self.pX,self.pY
#         elif 'x' in lab:
#             labs = ['$x(mm)$','$y(mm)$']
#             px,py = self.pxy
#             X,Y = self.pX*px*1e3,self.pY*py*1e3
#         return dsp.stddisp(im=[X,Y,image],labs=labs,
#             pOpt='ptX',title="%s" %os.path.basename(file),**kwargs)
#
# class Base_Viewer:
#     def __init__(self,figpath='',frame=None,thick=5,cutoff=50,i=0,v=1,pargs={}):#,fig=None,ax=None,):
#         self.fig,self.ax = dsp.stddisp()
#         cid = self.fig.canvas.mpl_connect('key_press_event', self)
#         self.figpath = figpath
#         self.nfigs   = self._get_nfigs()
#         if frame:
#             i = min(max(frame,0),self.nfigs)-1
#         else:
#             frame = i+1
#         self.frame  = frame       #starting image
#         self.i      = i           #starting image
#         self.inc    = 1           #increment(use 'p' or 'm' to change)
#         self.cutoff = cutoff
#         self.thick  = thick
#         self.fieldNames  = ['thick','frame','cutoff','inc']
#         self.mode = 1
#         self.pargs  = pargs
#         rc('text', usetex=False)
#         self.show()
#         if v:self.show_help()
#
#     def __call__(self, event):
#         # print(event.key)
#         if event.key in ['up','right']:
#             self.i=min(self.i+self.inc,self.nfigs-1)
#             self.mode=1
#         elif event.key in ['left','down']:
#             self.i=max(0,self.i-self.inc)
#             self.mode=-1
#         self.frame=self.i+1
#
#         if event.key=='s':
#             dsp.saveFig(self.figpath+'_%s.png' %str(self.i).zfill(3),ax=self.ax)
#
#         #increment rate
#         if event.key=='p':
#             self.inc=min(self.inc+1,100)        ;print('increment rate : %d' %self.inc)
#         elif event.key=='m':
#             self.inc=max(1,self.inc-1)          ;print('increment rate : %d' %self.inc)
#         elif event.key=='ctrl+r':
#             self.inc=1                          ;print('increment rate : %d' %self.inc)
#
#         #brightness
#         if event.key=='pageup':
#             self.cutoff=min(self.cutoff+5,500)  ;print('cutoff : %d' %self.cutoff)
#         elif event.key=='pagedown':
#             self.cutoff=max(1,self.cutoff-5)    ;print('cutoff : %d' %self.cutoff)
#         elif event.key=='r':
#             self.cutoff=50                      ;print('cutoff : %d' %self.cutoff)
#
#         #thickness
#         if event.key=='ctrl+t':
#             self.thick+=5                       ;print('thickness : %d' %self.thick)
#         elif event.key=='ctrl+T':
#             self.thick=max(self.thick-5,5)      ;print('thickness : %d' %self.thick)
#
#         if event.key=='h':self.show_help()
#         elif event.key=='enter':self.settings()
#         keys = self.call(event)
#
#         update_keys = keys+['enter','ctrl+t','ctrl+T','pageup','pagedown','r','left','right','down','up']
#         if event.key in update_keys:self.show()
#
#     def settings(self):
#         fieldValues = ['%d' %self.__dict__[f] for f in self.fieldNames]
#         fieldNames = self.fieldNames.copy()
#         self.get_fieldValues(fieldNames,fieldValues)
#         dict_fv = multenterbox("Change settings","settings", fieldValues,fieldNames)
#         if dict_fv:
#             for f in self.fieldNames:
#                 self.__dict__[f] = int(dict_fv[f])
#             self.set_fieldValues(dict_fv)
#         self.i = self.frame-1
#
#     def show(self):
#         tle = "frame %d, thickness=%d $\AA$" %(self.i+1,self.thick)
#         self.ax.cla()
#         self.get_im(fig=self.fig,ax=self.ax,title=tle,opt='',**self.pargs)
#         self.fig.canvas.draw()
#
#     def show_help(self):
#         msg = '''
#     'up or right'  : show frame+1
#     'down or left' : show frame-1
#     ##
#     'p' : increase increment rate
#     'm' : decrease increment rate
#     ##
#     'pageup'   : increae cutoff brightness
#     'pagedown' : decrease cutoff brightness
#     'r'        : reset cutoff brightness
#     ##
#     'ctrl+t' : increase thickness
#     'ctrl+T' : decrease thickness
#     ##
#     'enter' : change settings
#     's' : save image
#     'h' : show help
#         '''
#         print(colors.green+'shortcuts : '+colors.blue+msg+colors.black)
#     ###################################
#     ##### virtual functions
#     ###################################
#     def _get_nfigs(self):
#         print('base nfigs')
#         return 0
#     def get_im(self,**kwargs):
#         print('base get_im')
#     def get_fieldValues(self,fieldNames,fieldValues):return None
#     def set_fieldValues(self,dict_fv):return None
#     def call(self,event):return []
#
#
# class Frames_Viewer(Base_Viewer):
#     '''Viewer for pets'''
#     def __init__(self,pets,thick,Imag,kargs,**sargs):
#         self.pets   = pets
#         self.kargs  = kargs
#         super().__init__(thick=thick,cutoff=Imag,**sargs)
#
#     def get_fieldValues(self,fieldNames,fieldValues):
#         fieldNames+=list(self.kargs.keys())
#         fieldValues+=[str(f) for f in self.kargs.values()]
#     def set_fieldValues(self,dict_fv):
#         float_keys = np.setdiff1d(list(self.kargs.keys()),'opts')
#         for f in float_keys:
#             self.kargs[f] = eval(dict_fv[f])
#         self.kargs['opts']=dict_fv['opts']
#     def _get_nfigs(self):
#         return self.pets.nFrames
#     def get_im(self,**kwargs):
#         self.pets.show_frame(frame=self.i+1,thick=self.thick,Imag=self.cutoff,
#             **self.kargs,**kwargs)
#
#     def call(self,event):
#         chars = 'KkPhr'
#         keys = ['ctrl+K','ctrl+k', 'ctrl+H','ctrl+h','ctrl+g']
#
#         vals = [c in self.kargs['opts'] for c in chars]
#         for i,(c,k) in enumerate(zip(chars,keys)):
#             if event.key==k:
#                 vals[i] = not vals[i]
#
#         self.kargs['opts']='q'+''.join([c for c,val in zip(chars,vals) if val])
#
#         return keys
#
# class Viewer(Base_Viewer):
#     '''similar to adxv. Works with raw cbf/tiff'''
#     def __init__(self,exp_path=None,
#         Smax=0.025,Nmax=13,rot=203,pets=None,v=1,init='',**sargs):
#         """ View cbf files
#         - exp_path : path to images
#         - figpath : place to save the figures
#         - i : starting image
#         """
#         if isinstance(pets,str):pets = pt.Pets(pets)
#         if isinstance(pets,pt.Pets):exp_path = os.path.join(pets.path,'tiff')
#
#         d_fmt = {'cbf':self.load_cbf,'tiff':self.load_tif,'tif':self.load_tif}
#         self.supported_fmts = d_fmt.keys()
#
#         self.exp_path = os.path.realpath(exp_path)
#         self.fmt      = self.find_format(v)
#         self.figs     = np.sort(glob.glob(self.exp_path+'/*.%s' %self.fmt))#;print(self.figs)
#         self.load     = d_fmt[self.fmt]
#         self.pets     = pets
#         if self.pets:
#             vals  = ['center','refl','pred','boxes']
#             keys  = ['c','r','s','b']
#             alias = ['1','2','3','4']
#             self.show_d = dict(zip(keys,vals))
#             self.alias  = dict(zip(alias,keys))
#             self.show_opt = dict(zip(vals,[c in init for c in keys ]))
#             self.Smax,self.Nmax,self.rot=Smax,Nmax,rot
#         super().__init__(v=v,**sargs)
#
#     def call(self,event):
#         # k = event.key
#         if self.pets:
#             if event.key in self.alias.keys():event.key = self.alias[event.key]
#             if event.key in self.show_d.keys():
#                 key = self.show_d[event.key]
#                 self.show_opt[key] = not self.show_opt[key]
#
#             return list(self.show_d.keys())
#         else:
#             return []
#     def get_im(self,**kwargs):
#         fig = self.figs[self.i]
#         figname = os.path.basename(fig)
#         print(colors.yellow+fig+colors.black)
#         im = self.load(fig)
#         plts,scat,pp = [],[],[]
#         sargs = {'facecolor':'none','edgecolor':(0.7,0.7,0.15),'s':50,'marker':'o'}
#         if self.pets:
#             # df = self.rpl
#             df = self.pets.rpl
#             frame = self.i+1
#             if self.show_opt['refl']:
#                 rpl = df.loc[df.F==frame]
#                 plts+=[[rpl.rpx-0.5,rpl.rpy-0.5,'r+','']]
#             if self.show_opt['center']  :
#                 cen = self.pets.cen.iloc[self.i]
#                 plts += [[ cen.px-0.5,cen.py-0.5,'b+','']]
#             if self.show_opt['pred'] or self.show_opt['boxes']:
#                 px,py,I,hkl = self.pets.get_kin(frame,rot=self.rot,thick=self.thick,Smax=self.Smax,Nmax=self.Nmax,pixel=True)
#             if self.show_opt['pred']:
#                 scat  = [px,py]
#             if self.show_opt['boxes']:
#                 npx = 15
#                 pp = [dsp.Rectangle((px0-npx/2,py0-npx/2),npx,npx,facecolor='none',edgecolor='r') for px0,py0 in zip(px,py)]
#         dsp.stddisp(plts,patches=pp,scat=scat,im=[im],ms=20,sargs=sargs,xylims=[0,516,516,0],
#             cmap='gray',caxis=[0,self.cutoff],pOpt='tX',**kwargs)
#
#     def _get_nfigs(self):
#         return self.figs.size
#
#     ###############################################################
#     #### misc
#     ###############################################################
#     def find_format(self,v=1):
#         fmts = np.unique([f.split('.')[-1] for f in os.listdir(self.exp_path)])
#         fmts = [fmt for fmt in fmts if fmt in self.supported_fmts]
#         if not len(fmts):
#             raise Exception('no supported format found in %s. Supported formats :' %(self.exp_path),self.supported_fmts)
#         fmt = fmts[0]
#         if len(fmts)>1:
#             print('warning multiple formats found',fmts)
#             print('using %s' %fmt)
#         if v:print('%s format detected' %fmt)
#         return fmt
#
#     def load_cbf(self,fig):
#         try:
#             content = cbf.read(fig)
#         except:#UnicodeDecodeError
#             self.i=self.i+self.mode
#             print(colors.red+'error reading file'+colors.black)
#             self.import_exp()
#             return
#         numpy_array_with_data = content.data
#         header_metadata = content.metadata
#         # print(colors.blue,header_metadata,colors.black)
#         return numpy_array_with_data
#
#     def load_tif(self,fig):
#         return tifffile.imread(fig)
#
#
# def multenterbox(msg,title,fieldValues,fieldNames):
#     fieldValues = easygui.multenterbox(msg, title, fieldNames,fieldValues)
#     while True:
#         if fieldValues is None:
#             break
#         errs = list()
#         for n, v in zip(fieldNames, fieldValues):
#             if v.strip() == "":errs.append('"{}" is a required field.'.format(n))
#         if not len(errs):
#             break
#         fieldValues = easygui.multenterbox("\n".join(errs), title, fieldNames, fieldValues)
#     return dict(zip(fieldNames,fieldValues))
#
#


#
# ###########################################################################
# #### misc :
# ###########################################################################
def _get_cif_file(path,cif_file):
    if not cif_file:
        cif_file = _find_files(path,'cif')
    return cif_file


def _find_files(path,f_type):
    files = glob.glob(path+'/*.%s' %f_type)
    if len(files)==1:
        file=files[0]
        return file
    else:
        msg ='''
    only 1 %s file should be found in path folder :
     %s
    but %d were found :
     %s''' %(f_type,os.path.realpath(path),len(files),str(files))
        raise Exception(msg)

# def load_tif(fig):
#     return tifffile.imread(fig)
#
# def load_cbf(fig):
#     try:
#         content = cbf.read(fig)
#         numpy_array_with_data = content.data
#         header_metadata = content.metadata
#         return numpy_array_with_data
#     except:#UnicodeDecodeError
#         self.i=self.i+self.mode
#         print(colors.red+'error reading file'+colors.black)

def find_format(exp_path,v=1):
    d_fmt = {'cbf':load_cbf,'tiff':load_tif,'tif':load_tif}
    supported_fmts = d_fmt.keys()

    fmts = np.unique([f.split('.')[-1] for f in os.listdir(exp_path)])
    fmts = [fmt for fmt in fmts if fmt in supported_fmts]
    if not len(fmts):
        raise Exception('no supported format found in %s. Supported formats :' %(exp_path),supported_fmts)
    fmt = fmts[0]
    if len(fmts)>1:
        print('warning multiple formats found',fmts)
        print('using %s' %fmt)
    if v:print('%s format detected' %fmt)
    return fmt,d_fmt[fmt]
#
#
# def import_crys(file):
#     if file.split('.')[-1]=='cif':
#         crys = Crystal.from_cif(file)
#     elif sum(np.array(list(Crystal.builtins))==file):
#         crys = Crystal.from_database(file)
#     else:
#         raise Exception('cannot import %s' %file)
#     return crys



# -*- coding: utf-8 -*-
# %% 0: Subroutines
"""
Created on Mon Jul 12 17:16:01 2021

@author: Richard

Import data from a reflprofiles PETS file
"""

import numpy as np
import matplotlib.pyplot as plt
import os
from gemmi import cif


def strucfac(hkl, B):
    # atom coordinates
    aSi = 5.432
    r = np.array([0.0, 0.0, 0.0,
                  0.5, 0.5, 0.0,
                  0.0, 0.5, 0.5,
                  0.5, 0.0, 0.5,
                  0.25, 0.25, 0.25,
                  0.75, 0.75, 0.25,
                  0.25, 0.75, 0.75,
                  0.75, 0.25, 0.75])
    r = r.reshape(8, 3)
    gmag = (1/aSi)*np.dot(hkl, hkl)**0.5
    # calculate scattering factor
    aSi = ([1.065438920E+00, 1.201436910E-01, 1.809152630E-01])
    bSi = ([1.041184550E+00, 6.871133680E+01, 8.875339260E-02])
    cSi = ([1.120656200E+00, 3.054528160E-02, 1.599635020E+00])
    dSi = ([3.700626190E+00, 2.140978970E-01, 9.990966380E+00])
    fSi = 0
    for i in range(3):
        fSi = fSi + aSi[i]/((gmag**2)+bSi[i])
        + cSi[i]*np.exp(-(dSi[i]*gmag**2))

    # structure factor
    F = np.sum(fSi*np.exp(-B*gmag*gmag/4)*np.exp(2*np.pi*1j*np.dot(r, hkl)))
    return(F)


def writefelixinp(u, v, w, x1, x2, x3, npix, lpha, th_start, th_end, th_step):
    # currently this just writes a basic Felix file
    # but expect more input arguments to be added

    fel = open('felix.inp', 'w')

    fel.write('# Input file for Felix version :master: v1.03 :14Mar2019:\n')
    fel.write('# ------------------------------------\n\n')
    fel.write('# ------------------------------------\n\n')
    fel.write('# control flags\n')
    fel.write('IWriteFLAG                = 1\n')
    fel.write('IImageFLAG                = 1\n')
    fel.write('IScatterFactorMethodFLAG  = 0\n')
    fel.write('IBlochMethodFLAG          = 0\n')
    fel.write('IMaskFLAG                 = 1\n')
    fel.write('IHolzFLAG                 = 0\n')
    fel.write('IAbsorbFLAG               = 2\n')
    fel.write('IAnisoDebyeWallerFlag     = 0\n')
    fel.write('IByteSize                 = 8\n\n')
    fel.write('# radius of the beam in pixels\n')
    fel.write('IPixelCount               = '+npix+'\n\n')
    fel.write('# beam selection criteria\n')
    fel.write('IMinReflectionPool        = 200\n')
    fel.write('IMinStrongBeams           = 100\n')
    fel.write('IMinWeakBeams             = 0\n\n')
    fel.write('# crystal settings\n')
    fel.write('RDebyeWallerConstant      = 0.0\n')
    fel.write('RAbsorptionPer            = 10.0\n\n')
    fel.write('# microscope settings\n')
    fel.write('ROuterConvergenceAngle    = '+lpha+'\n')
    # Incident beam - from the input
    fel.write('IIncidentBeamDirection    = ['+u+','+v+','+w+']\n')
    # x-direction as calculated from cross product
    fel.write('IXDirection               = ['+x1+','+x2+','+x3+']\n')
    # not very sure about this - assume it is the beam direction
    # fel.write('INormalDirection          = ['+u+','+v+','+w+']\n')
    fel.write('INormalDirection          = [-1,0,1]\n')
    fel.write('RAcceleratingVoltage (kV) = 200.0\n')
    fel.write('RAcceptanceAngle          = 0.0\n\n')
    fel.write('# Image Output Options\n')
    fel.write('RInitialThickness        = '+str(th_start)+'\n')
    fel.write('RFinalThickness          = '+str(th_end)+'\n')
    fel.write('RDeltaThickness          = '+str(th_step)+'\n')
    fel.write('RPrecision               = 0.00002\n\n')
    fel.write('#Refinement Specific Flags\n')
    fel.write('IRefineModeFLAG          = S\n')  # simulation only
    # all these lines are irrelevant for sim only
    fel.write('IWeightingFLAG           = 0\n')
    fel.write('IRefineMethodFLAG        = 3\n')
    fel.write('ICorrelationFLAG         = 2\n')
    fel.write('IImageProcessingFLAG     = 4\n')
    fel.write('RBlurRadius              = 0.0\n')
    fel.write('INoofUgs                 = 40\n')
    fel.write('IAtomicSites             = (1)\n')
    fel.write('IPrint                   = 0\n')
    fel.write('RSimplexLengthScale      = 0.1\n')
    fel.write('RExitCriteria            = 0.000001\n')

    fel.close()


def hklfile(xtal, hkl):
    # make a filename corresponding to a felix output e.g. Si_+0-2+7.bin
    if (hkl[0] < 0):
        h_str = str(hkl[0])
    else:
        h_str = '+'+str(hkl[0])
    if (hkl[1] < 0):
        k_str = str(hkl[1])
    else:
        k_str = '+'+str(hkl[1])
    if (hkl[2] < 0):
        l_str = str(hkl[2])
    else:
        l_str = '+'+str(hkl[2])
    fname = xtal + '_' + h_str+k_str+l_str + '.bin'

    return fname


def gaussian(r, sigma):
    return np.exp(-(r*r/(2*sigma)))


def conv2D(imagein, kernel):
    wx = len(imagein)
    wc = len(kernel)
    cc = int(wc/2)
    imgout = imagein*0
    for i in range(wc):
        for j in range(wc):
            imgout[i:wx-wc+i, j:wx-wc+j] += imagein[cc:wx+cc-wc,
                                                    cc:wx+cc-wc]*kernel[i, j]
    return imgout


def optimise_orientation(xtal,sim, local_hkl_list, local_fmax_list, inpath, outpath,
                         fram, kernel, plot, save):
    # Get the peak positions (frame number) for reflections in a given LACBED
    # pattern and vary the x-coordinate of the origin so that the peaks
    # lie along a line parallel to x
    # number of DF LACBED patterns
    if(plot):
        print('Simulation:%d' %sim)
    # os.chdir(inpath)
    start = -9+inpath[-10:].find('x')
    wx = int(inpath[start:])  # the patterns are wx x wx pixels
    sim_lacbed = ([])
    # load the LACBED patterns for this simulation
    for hkl in local_hkl_list:
        fnom = hklfile(xtal, hkl)
        lacbed_name = os.path.join(inpath,fnom)
        print(lacbed_name)
        if(os.path.exists(lacbed_name)):
            if '+0+0+0' not in lacbed_name:
                sim_lacbed.append(conv2D(
                    np.fromfile(lacbed_name, dtype=np.float64)
                    .reshape(wx, wx), kernel))
        else:
            print(fnom+' not found')

    # now find out if the offset in the frame indexing and the best line
    # that matches the relative positions of the maxima in different frames.
    # We assume that the x-direction is correct.
    # The calculation is quick so just go through a range and find the
    # one which gives the closest to a straight line along x
    fom_list = ([])
    best_fom = 1.0e+10  # a big number, we want a minimum
    frame_range = 10  # how far we look for the origin +/-
    # NB first frame starts at x=wx/4, each frame covers wx/(2*fram) pixels
    # number of first frame is 2*sim*fram +1 (e.g. fram=10 gives 1,21,41...)
    frf = frame_range*fram
    for offset in range(-frf, frf):
        row_list = ([])
        # if(plot):
        #     plt.figure()
        #     plt.title(str(offset))
        # get the figure of merit - i.e. the sum of squares of
        # deviations of peak positions from a horizontal line
        # looping over the reflections in this simulation
        for j in range(len(sim_lacbed)):
            # pixel coords of the maximum for this reflection
            # we have sub-frame accuracy since fmax is a centroid
            col = int(wx/4 + fram*(local_fmax_list[j] - 2*sim*fram)+offset+0.5)
            # print(str(offset)+', '+str(j)+'; max '+str(local_fmax_list[j])
            # +' =column '+str(col)+'\n')
            # intensity for the specified column
            I_sim = sim_lacbed[j][:, col]
            # if(plot):
            #     plt.plot(I_sim)
            # the row corresponding to the centroid intensity of this column
            # if PETS gave a perfect answer this would be the central row of
            # the simulated image for all reflections
            row_list.append(np.dot(list(range(wx)), I_sim)/np.sum(I_sim))
        # if(plot) & (save):
        #     plotname = outpath + '\\' + str(sim) + '_' +str(offset)+ '.png'
        #     plt.savefig(plotname)
        #     plt.close()
        # find the sum of squares of the deviation from the average row
        # this is our figure of merit fom_row
        mean_row = np.mean(row_list)
        fom_row = np.dot((row_list-mean_row), (row_list-mean_row))
        if (fom_row < best_fom):
            best_fom = fom_row
            best_row = int(np.rint(mean_row))
            best_offset = offset/fram
        fom_list.append(fom_row)
    # plot if required
    if(plot):
        plt.figure()
        plt.plot(np.arange(-frame_range, frame_range, 1/fram), fom_list)
        plt.title(str(sim))
        plt.xlabel('Frame offset')
        plt.ylabel('Figure of merit')
        # plt.show
        print('best row = '+str(best_row)+', offset = '
              + str(best_offset) + ' frames')
    # save plot if required
    if(save):
        plotname = os.path.join(outpath,str(sim) + '.png')
        plt.savefig(plotname)
        plt.close()

    return best_row, best_offset



''' Computes path length for all pixel in a detector with respect 
to :
    Incident beam with potentially spatial and spectral broadening
    Atomic distribution with potentially non zero B factor
features :
    - Fraunhofer diffraction framework
    - Loss of coherency option
'''
#nearBragg program
#main                                                   #58
#parse arguments                                        #150
#display help                                           #400
#allocate                                               #450
#handle parameters                                      #500
#load lattice points                                    #750
#initializing                                           #815

#The mega loop
for yp,xp in zip(ypixels,xpixels):                      #950
    Fa,Fb = 0,0
    for suby,subz in zip(oversample,oversample):        #976
        pixel=[distance,Ydet-Ybeam,Xdet-Xbeam]
        steps = divsteps*dispsteps*depthsteps*oversample*oversample
        for lam in wavelengths:                         #1010
            for hdiv,vdiv in solid_angle:               #1015
                for source_tic in source_pts:           #1024
                    for i in atoms:                     #1042
                        phase=2*pi*(source_to_atom_path+atom_to_pixel_path)/lam + phsft
                        S0,S = atom-source, pixel-atom
                        stol=norm(S-S0)/lam*1e10
                        DWF = occ[i]*exp(-Bfac[i]*stol**2)
                        Fa+=DWF*cos(phase)/source_to_atom_path/atom_to_pixel_path
                        Fb+=DWF*sin(phase)/source_to_atom_path/atom_to_pixel_path
                if not coherent : I += Fa**2 + Fb**2; Fa,Fb=0,0
            if not coherent : I += Fa**2 + Fb**2; Fa,Fb=0,0
    if coherent :                                       #1120
        Fa += cosimage[j]*steps
        Fb += sinimage[j]*steps
        cosimage[j] = Fa/steps
        sinimage[j] = Fb/steps
        Fa /= sqrt(steps);
        Fb /= sqrt(steps);
    I += Fa**2 + Fb**2
    floatimage[j] += I/steps*omega_Rsqr_pixel*fluence*r_e_sqr
    j+=1

# write floatimage into 'floatimage.bin'                #1193
# convert and dump into file 'intimage.img'             #1211
intimage = float2gray(floatimage)
np.savetxt('intimage.img',intimage)



from utils import*
from scipy.integrate import trapz,quad
from scattering_factors import*
fig_path = get_figpath(__file__,'/figures/')


def display_MottBethe_fit(qmax=3,npts=100,**kwargs):
    elts = ['C']
    q,fq_x = get_xray_atomic_factors(elts,qmax,npts)
    q,fq_e = get_elec_atomic_factors(elts,qmax,npts)

    Z,q2=6, q**2
    fq_x,fq_e=fq_x[0],fq_e[0]
    f_MB = 1/(2*pi**2*a0)*(Z-fq_x)/q2
    #f_z = 1/(2*pi**2*a0)*Z/q2
    #fe0 = Z/(3*a0)*r2

    plts = [[q,fq_e,'b','$Gauss-fit$'],
            [q,f_MB,'r*','$Mott-Bethe$'],
            #[q,f_z**2,[(1,0.7,0.7),'--'],'$Z/q^2$']
            ]
    stddisp(plts,labs=['$q(A^{-1})$','$f^e(A)$'],xylims=[0,qmax,0,3],**kwargs)

# def compute_lenz_int(Z=6,KE=100):
#     v,lam = get_v_from_KE(KE),wavelength(KE)
#     g2,k0 = 1/(1-v**2),2*pi/lam
#     theta0 = pow(Z,1/3)/(k0*a0)
#
#     t = np.linspace(0,pi,100000)
#     fq_l = lambda t:2*np.sin(t)/(t**2+theta0**2)**2
#     si=trapz(fq_l(t),dx=t[1]-t[0])
#     siq=quad(fq_l,0,pi)
#     print(si,siq[0],1/theta0**2, theta0)
#     return siq
#
# def display_Lenz():
#     KE,Zmax=100,20
#     Z,v = list(range(1,Zmax+1)),get_v_from_KE(KE)
#     sigma_Lenz = 1.87e-4*np.array(Z)**(4/3)/v**2
#     plts = [#[Z,sigma,'rs-','$th$'],
#             [Z,sigma_Lenz,'bs--','$Lenz$']]
#     stddisp(plts,labs=['$Z$','$\sigma_e(A^2)$'],opt='p',gridmOn=True,logOpt='xy')



display_MottBethe_fit(qmax=3,npts=100,lw=3,legLoc='upper right',title='Carbone'
    name=fig_path++'MottBethe.svg',opt='p')



import importlib as imp
import os,glob, numpy as np, pandas as pd, tifffile,mrcfile,scipy.optimize as opt
from typing import TYPE_CHECKING, Dict, Iterable, Optional, Sequence, Union
from subprocess import Popen,PIPE,check_output
from crystals import Crystal,Lattice
from utils import handler3D as h3D          #;imp.reload(h3D)
from utils import displayStandards as dsp   #;imp.reload(dsp)
from utils import glob_colors as colors
from multislice  import mupy_utils as mut   #;imp.reload(mut)
from EDutils import viewers as vw           #;imp.reload(vw)
from EDutils import utilities as ut         #;imp.reload(ut)
from multislice.rotating_crystal import get_crystal_rotation
from gemmi import cif
from utils import physicsConstants as cst
from blochwave import bloch                 #;imp.reload(bloch)
# from . import import_ED as ED               ;imp.reload(ED)


class Pets():
    def __init__(self,pts_file:str='',path:str='',
        cif_file:Optional[str]=None,gen:bool=False,dyn:bool=False):
        """ Pets importer

        parameters
        ----------
            pts_file
                full path to .pts file or .cif_pets
            cif_file
                full path to .cif file (automatically found if None)
            gen
                force reload if True
            dyn
                load dyn_cif into .cif if True
        """
        if path:
            # print(glob.glob(os.path.join(path,'*.pts')))
            pts_file = glob.glob(os.path.join(path,'*.pts'))[0]
        self.name = os.path.basename(pts_file).split('.pts')[0]
        self.path = os.path.dirname(pts_file)
        self.out  = self.path+'/dat/'
        # print('path %s,pts_file : %s ' %(self.path,self.name))


        self.cif_file   = ut.find_cif_file(self.path,cif_file)
        self.crys       = ut.import_crys(self.cif_file)
        self.lat_vec    = np.array(self.crys.lattice_vectors)
        self.lat_vec1   = np.array(self.crys.reciprocal_vectors)/(2*np.pi)
        self.lat_params = self.crys.lattice_parameters

        gen |= not os.path.exists(self.out)
        # if os.path.exists(self.out):gen |= not os.path.exists(self.out+'rpl.txt')
        if gen:self._convert_pets()
        self._load_all(dyn)
        self.nFrames = self.uvw.shape[0]

        tiffs = glob.glob(os.path.join(self.path,'tiff','*.tiff'))
        if len(tiffs):
            self.nxy = tifffile.imread(tiffs[0]).shape
        else:
            self.nxy = np.round([self.rpl.px.max(),self.rpl.py.max()])

    def hkl_to_pixels(self,h,frame):
        hkl = np.array([eval(hkl) for hkl in h])
        #spot locations in reciprocal crystal frame (with shape 3 x n_refl)
        rxyz = self.UB.dot(hkl.T).T
        # print(rxyz)
        alpha,beta,gamma = self.cif.iloc[frame-1][['alpha','beta','omega']]
        # print(alpha,beta,gamma)
        alpha_r,beta_r,gamma_r = -np.deg2rad([alpha,beta,gamma])
        ctx,stx = np.cos(alpha_r),np.sin(alpha_r)
        cty,sty = np.cos(beta_r) ,np.sin(beta_r)
        ctz,stz = np.cos(gamma_r),np.sin(gamma_r)

        Rx = np.array([[1,0,0],[0,ctx,stx],[0,-stx,ctx]])
        Ry = np.array([[cty,0,sty],[0,1,0],[-sty,0,cty]])
        Rz = np.array([[ctz,stz,0],[-stz,ctz,0],[0,0,1]])
        # R = Ry.dot(Rx.dot(Rz))
        R = Rz.dot(Ry.dot(Rx))
        qxyz = R.dot(rxyz.T)
        # print(qxyz)
        qx,qy = qxyz[:2,:]

        ### convert to pixels
        df_pxy=pd.DataFrame()
        cx,cy = self.cen.loc[frame-1, ['px','py']]
        df_pxy['px'] = qx/self.aper + cx
        df_pxy['py'] = -qy/self.aper + cy
        df_pxy.index=h
        return df_pxy

    def save(self):
        ut.save_pkl(self, os.path.join(self.path,'pets.pkl'))

    def _convert_pets(self):
        cmd = "%s/convert_pets.sh %s %s" %(os.path.dirname(__file__),
            self.path,self.name)
        # print(cmd)
        p = Popen(cmd,shell=True,stdout=PIPE,stderr=PIPE);p.wait();o,e=p.communicate()
        print(o.decode(),e.decode())

        A = np.loadtxt(self.out+'UB.txt')
        np.save(self.out+'UB.npy',np.reshape(A,(3,3)))

    def _add_hkl(self,df):
        hkl = self.invA.dot(df[['x','y','z']].values.T)
        df[['hx','kx','lx']]  = hkl.T
        df[['h','k','l']]     = np.array(np.round(hkl),dtype=int).T

    def _load_all(self,dyn=1):
        lam,omega,aper = np.loadtxt(self.out+'pts.txt')
        self.omega = omega
        self.aper  = aper
        self.lam   = lam
        self.K0    = 1/self.lam
        self.keV = cst.lam2keV(self.lam)

        self.frames = pd.read_csv(self.out+'iml.txt',sep=',',names=['name','alpha','beta','domega','scale','calibration','ellipA','ellipP','used'])
        self.rpl = pd.read_csv(self.out+'rpl.txt',sep=',',names=['x','y','z','I','i','px','py','rpx','rpy','alpha','Im','F'])
        self.cor = pd.read_csv(self.out+'cor.txt',sep=',',names=['x','y','z','I','i','px','py','rpx','rpy','alpha','Im','F'])
        self.cen = pd.read_csv(self.out+'cenloc.txt',sep=',',names=['px','py','m','n'])
        self.xyz = pd.read_csv(self.out+'xyz.txt',sep=',',names=['x','y','z','I','u0','px','py','F','alpha','Im','u1'])
        self.hkl = pd.read_csv(self.out+'hkl.txt',sep=',',names=['h','k','l','I','i','F','L'])
        self.kin = pd.read_csv(self.out+'cif.txt',sep=',',names=['id','u','v','w','prec','alpha','beta','omega','scale'])
        self.dyn = pd.read_csv(self.out+'dyn.txt',sep=',',names=['id','u','v','w','prec','alpha','beta','omega','scale'])
        self.HKL = pd.read_csv(self.out+'HKL.txt',sep=',',names=['h','k','l','I','sig','F'])
        self.HKL_dyn = pd.read_csv(self.out+'HKL_dyn.txt',sep=',',names=['h','k','l','I','sig','F'])
        self.A   = np.load(self.out+'UB.npy')
        self.UB  = self.A
        self.lat_params = np.loadtxt(self.out+'cell.txt')[:-1]
        self.lat = np.array(Lattice.from_parameters(*self.lat_params).lattice_vectors)
        self.invA = np.linalg.inv(self.A)

        self.cif = [self.kin,self.dyn][dyn]
        uvw   = self.cif[['u','v','w']].values

        beams = self.lat.T.dot(uvw.T).T

        self.uvw   = uvw #/np.linalg.norm(uvw,axis=1)[:,None]
        self.uvw0  = -beams/np.linalg.norm(beams,axis=1)[:,None]
        self.beams = self.K0*self.uvw0 #/np.linalg.norm(beams,axis=1)
        self.XYZ   = self.xyz[['x','y','z']].values.T

        self._add_hkl(self.rpl)
        self._add_hkl(self.cor)
        self._add_hkl(self.xyz)

        cx,cy = self.cen[['px','py']].iloc[self.rpl.F-1].values.T
        px,py = self.rpl[['px','py']].values.T
        qxqy  = self.aper*(np.vstack([px,-py]).T-np.vstack([cx,-cy]).T)
        self.rpl[['qx','qy']] = qxqy
        self.rpl['hkl'] = [str(tuple(h)) for h in self.rpl[['h','k','l']].values]

        self.alpha = self.cif.alpha.values
        self.hkl.index=[str(tuple(h)) for h in self.hkl[['h','k','l']].values]
        hkl = self.hkl[['h','k','l']].values
        rq = np.linalg.norm(hkl.dot(self.lat_vec1),axis=1)
        self.hkl['rq'] = rq

        # hkl = [str(tuple(h)) for h in self.xyz[['h','k','l']].values]
        # hkl0,idx,cc=np.unique(hkl,return_index=True,return_counts=True)
        # if not len(hkl)==idx.shape[0]:print('warning reflection not unique : ',hkl0[cc>1])
        # self.xyz.index=hkl

        # beams = self.lat.T.dot(dyn.uvw.T).T
        # self.dyn[['u0','v0','w0']] = beams/np.linalg.norm(beams,axis=1)[:,None]

        if dyn:
            hkl,idx=np.unique([str(tuple(h)) for h in self.HKL_dyn[['h','k','l']].values],return_index=True)
            self.HKL_dyn=self.HKL_dyn.iloc[idx]
            self.HKL_dyn.index=hkl
            hkl = self.HKL_dyn[['h','k','l']].values
            self.HKL_dyn['rq'] = np.linalg.norm(hkl.dot(self.lat_vec1),axis=1)


    def load_b0(self):
        return ut.load_pkl(self.b0_path)

    def run_blochwave(self,F,**bargs):
        Nmax = self.HKL_dyn[['h','k','l']].max().max()
        self.dyn.set_index('id')
        u = self.dyn.loc[F,['u','v','w']].values

        u0 = self.lat_vec.T.dot(u)
        u0 /= np.linalg.norm(u0)

        b_args=dict(Nmax=Nmax,Smax=0.05,path=self.dyngo_path,
            opts='',name='b_F%d' %F,solve=True)
        b_args.update(bargs)
        b0 = bloch.Bloch(self.cif_file,keV=200,
            u=-u0,**b_args)
        self.b0_path=b0._get_pkl()
        return b0

    def make_eldyn(self,
            sgmax=1,sslim=0.01, outputsgmax=0.1,
            outputprecfrac=0.6,phisteps=50,F=1,
            thickness=920.9828,F0=0,bargs={},
            thr=8,path=None,
            v=0):
        ''' Produces a <structure>.eldyn file
        '''

        UB        = self.A
        lam       = self.lam
        pets_dat  = self.dyn.loc[F]
        hklz      = pets_dat[['u','v','w']].values
        alpha,beta,omega,scale=pets_dat[['alpha','beta','omega','scale']]
        phi       = self.dyn.loc[F+1,'alpha']-alpha
        xnorm, ynorm, keys = 0,0,1000

        #### reflections
        self.HKL_dyn.index = [str(tuple(h)) for h in self.HKL_dyn[['h','k','l']].values]
        # hkls = self.HKL_dyn[['h','k','l','I','sig']].copy()
        # Nmax = self.HKL_dyn[['h','k','l']].max().max()

        structure = os.path.dirname(self.cif_file)[:-4]
        if not isinstance(path,str):
            path = os.path.join(self.path,'dyngo')
        if not os.path.exists(path):os.mkdir(path)
        eldyn_file=os.path.join(path,'%s.eldyn' %structure)
        self.dyngo_path=path
        ###
        if v:print('...getting structure factor from bloch module...')
        b0=self.run_blochwave(F,**bargs)
        # b_args=dict(Nmax=Nmax,Smax=0.05,path=path,opts='',
        #     name='b_F%d' %F,solve=True)
        # b_args.update(bargs)
        # b0 = bloch.Bloch(self.cif_file,keV=200,
        #     u=-hklz,**b_args)
        # self.b0_path=b0._get_pkl()

        # Fhkl = b0.Fhkl
        # V0_idx = np.array([2*Nmax]*3)
        # Fhkl[tuple(V0_idx)] = 0
        print('...getting all hkls...')
        hkls = pd.DataFrame(columns=['h','k','l','A','B','F','I','sig'],
            index=[ str(tuple(hkl)) for hkl in np.vstack([ i.flatten() for i in b0.hklF]).T],
            )

        Fhkl = b0.Fhkl.flatten()
        # Fhkl = np.array([ Fhkl[tuple(hkl_G+V0_idx)]
        #     for hkl_G in hkls[['h','k','l']].values ])
        hkls[['h','k','l']] = np.vstack([ i.flatten() for i in b0.hklF]).T
        hkls['A'] = np.array(Fhkl.real,dtype=float)
        hkls['B'] = np.array(Fhkl.imag,dtype=float)
        hkls['F'] = 6
        hkls['I']   = 0
        hkls['sig'] = 0
        # get the reflections to compute the intensities
        hkl_dyn = self.HKL_dyn[self.HKL_dyn.F==F]
        idx = hkl_dyn.sort_values('I')[-4:].index ##debug
        hkls.loc[idx,'F']         = 5
        hkls.loc[idx,['I','sig']] = hkl_dyn.loc[idx,['I','sig']].values
        # hkls.drop(str((0,0,0)))
        hkls = hkls.drop(str((0,0,0)))
        # print(hkls.loc[str((0,0,0))])
        print(hkls.loc[idx,'I'])

        fmt0 = lambda x,s,n:(('%'+str(s)+'f') %x).rjust(n)
        fmt  = lambda x:fmt0(x,6,12)

        # hkls=[
        #     [1,-2,-1, 0.1,0.2,0.3,0.4, 3,0.1],
        #     [-3,4,-2, 0.1,0.4,0.5,0.1, 2,0.1],
        #     ]
        # df_hkl = pd.DataFrame(hkls,columns=['h','k','l','I1','sig','re','im','d','e'])


        header="""int iedt thr %d

Zone# %d
Noncentrosymmetric         0
Refinement I
"""%(thr,F)

        ub='\n'.join(['%s%s%s' %(
            ('%.6f' %v[0]).rjust(12),
            ('%.6f' %v[1]).rjust(12),
            ('%.6f' %v[2]).rjust(12),
                )
            for v in UB])

        params = ''.join(fmt(s) for s in
            [lam,F0,omega,sgmax,sslim, outputsgmax,
            outputprecfrac]
            ) + ' %d' %phisteps

        fmt9 = lambda x:fmt0(x,6,9)
        geo = ''.join([fmt9(s) for s in hklz.tolist()+[alpha,beta,phi]])
        # '%s %s %s' %(
        #     ''.join(['%.6f' %x for x in hklz]),
        #     ''.join(['%.6f' %x for x in [alpha,beta] ]),
        #     '%.6f' %gamma,
        #     )

        thick =' '+' '.join([('%1.6f' %x)[:8] for x in [ scale, thickness,xnorm, ynorm]]) + ' '*24 + '%-4d' %keys
        cor = ' 0.000000 0.000000                                          00'

        msg=''
        msg+= header
        msg+= ub        +'\n'
        msg+= params    +'\n'
        msg+= geo       +'\n'
        msg+= thick     +'\n'
        msg+= cor       +'\n'
        # msg+=df_hkl.to_csv(float_format='%15.5E',sep='\t',index=False,header=False,line_terminator="\n\n\n")
        for i,hkl in hkls.iterrows():
            str_refl = '%s%s%s\n\n\n' %(
                ''.join(['%4d'    %x for x in hkl[['h','k','l']]]),
                ''.join(['%15.5E' %x for x in hkl[['I','sig','A','B']]]),
                ('%d' %hkl['F']).rjust(5) )
            msg+=str_refl
        with open(eldyn_file,'w') as f:f.write(msg)
        if v:
            print(colors.yellow+eldyn_file+colors.green+" saved"+colors.black)
            with open(eldyn_file,'r') as f:print(f.read())
        self.eldyn_file=eldyn_file
        return eldyn_file

    def run_dyngo(self,bin='dyngo/dyn'):
        print(check_output('%s %s' %(bin,self.eldyn_file),
            shell=True).decode())

    def get_edout(self,F=1,opts='a',v=0):
        #reformat output properly into a file
        edout = self.eldyn_file.replace('eldyn','edout')
        out = edout.replace('edout','out')
        cmd = '''\
    tail --lines=+2  %s | \
    sed -E "s/ {1,}/ /g" | sed 's/^ //' | \
    awk '{getline b; getline c;printf("%%s %%s %%s\\n",$0,b,c)}' |\
    sed "s/ /,/g" | sed "s/,$//" > %s''' %(edout,out)
        # print(cmd)
        print(check_output(cmd,shell=True).decode())

        #import data
        df = pd.read_csv(out,sep=',',names=['h','k','l','I'] + ['I_%d' %i for i in range(6)],
            index_col=None)

        print(check_output('rm %s' %out,shell=True).decode())
        # df[['h','k','l']]=np.array(df[['h','k','l']].values,dtype=int)
        df.index=[str(tuple(h)) for h in df[['h','k','l']].values]

        if 'a' in opts:
            df_allint=self.read_allint(F)
            df['Sw']=np.nan
            df.loc[df_allint.index,'Sw']=df_allint.Sw

        self.HKL_dyn.loc[df.index,'Icalc'] = df.I
        if v:
            print(self.HKL_dyn.loc[df.index,['I','Icalc']])
        self.save()
        return df

    def read_allint(self,F,opt=''):
        '''excitation errors from dyngo'''
        all_int =self.eldyn_file.replace('.eldyn','_%s.allint' %str(F).zfill(3))
        if os.path.exists(all_int) :
            ##the ones that made it to the int>deps
            df_allint = pd.read_csv(all_int,sep=' +',
                names=['h','k','l','I','S1','S2','Sw','F'],engine='python')
            df_allint.index=[str(tuple(h)) for h in df_allint[['h','k','l']].values]
            if not 'O' in opt:df_allint=df_allint.drop(str((0,0,0)))
            return df_allint

    ###########################################################################
    #### compute :
    ###########################################################################
    def integrate_rpl(self,frames,cond='(I>10)',npx=10):
        """Perform manual integration"""
        if isinstance(frames,int):frames=np.array(frames)
        cond += ' & (F in %s) ' %str(list(frames))
        cond += ' & (px>%d) & (py>%d) & (px<%d) & (py<%d)' %tuple([npx]*2+[512-npx]*2)
        rpl = self.rpl.loc[self.rpl.eval(cond)]
        # print(rpl.shape)
        # pxy = rpl[['px','py']].values
        x0 = np.arange(-npx,npx+1)
        x,y = np.meshgrid(x0,x0)
        x1,y1 = np.meshgrid(np.arange(-npx,npx+1,0.1),np.arange(-npx,npx+1,0.1))

        tiffpath = os.path.join(self.path,'tiff')
        df=pd.DataFrame(columns=['Imax','px','py','sx','sy','noise','max_err','mean_err','min_err'])
        for i,frame in enumerate(frames):
            print(frame)
            im = tifffile.imread(os.path.join(tiffpath, '%s.tiff' %str(frame).zfill(5)))
            # dsp.stddisp(im=[im],pOpt='im',caxis=[0,50],cmap='gray')
            for j,r in rpl.loc[rpl.F==frame].iterrows():
                px,py = int(r.px),int(r.py)
                data  = im[py-npx:py+npx+1,px-npx:px+npx+1]

                p0 = (r.I,0,0,5,5,r.i)
                popt, pcov = opt.curve_fit(f_gauss2D, (x, y), data.ravel(), p0=p0)
                # print('I=%.1f,x0=%.1f,y0=%.1f,sx=%.1f,sy=%.1f,noise = %.1f' %tuple(popt))
                # dsp.stddisp(im=[data],pOpt='im',cmap='gray',caxis=[0,50])#;dsp.plt.show()
                # fig,ax = dsp.stddisp(im=[x,y,data],cmap='gray',
                #     contour=[x1,y1,gauss2D((x1,y1),*popt),8],cargs={'colors':'r'},pOpt='im',lw=2)
                # fit = gauss2D((x,y),*popt)
                # dsp.stddisp(im=[fit],pOpt='im',cmap='gray',caxis=[0,50])#;dsp.plt.show()
                # dsp.plt.show()
                err = abs(f_gauss2D((x,y),*popt)-data.ravel())
                hkl = str(tuple(r[['h','k','l']].values))+'_%d' %frame
                df.loc[hkl] = list(popt)+[err.max(),err.mean(),err.min()]
        df['Im'] = rpl.Im.values
        df['I']  = rpl.I.values
        df['i']  = rpl.i.values
        df['F']  = rpl.F.values
        df['Iint'] = df.Imax*df.sx*df.sy*np.pi*self.aper
        return df

    ###########################################################################
    #### get
    ###########################################################################
    def get_lattice(self,Nmax=5):
        return mut.get_lattice(self.lat_vec1,Nmax)

    def get_beam_dir(self,frame):
        return self.uvw[frame-1]

    def get_beam(self,frame):
        return self.beams[frame-1]

    def get_ewald(self,frame,nts=100,nps=200):
        return mut.get_ewald(self.get_beam(frame),nts,nps)

    def get_excitation_errors(self,frame,Nmax=5,Smax=0.02):
        return mut.get_excitation_errors(self.get_beam(frame),self.lat_vec1,Nmax=Nmax,Smax=Smax)

    def get_kin(self,frame,thick,Nmax=5,Smax=0.02,e0=[1,0,0],rot=0,Imag=10,pixel=False):
        """get kinematic intensities"""
        K = self.get_beam(frame)#;print(K)
        df = mut.get_kinematic_intensities(self.cif_file,K,thick,Nmax=Nmax,Smax=Smax)
        qxyz = df[['qx','qy','qz']].values
        I = df.I.values*Imag

        px,py = mut.project_beams(K,qxyz,e0)
        if rot:
            ct,st = np.cos(np.deg2rad(rot)),np.sin(np.deg2rad(rot))
            px,py = ct*px-st*py,st*px+ct*py
        hkl = df[['h','k','l']].values.T
        if pixel:
            cx,cy = self.cen.loc[frame-1,['px','py']].values
            aper  = self.aper
            px= px/aper + cx
            py=-py/aper + cy                                        #;print(px,py)

        return px,py,I,hkl

    def get_hklI(self,refl):
        return self.hkl.loc[refl,'I'].values

    # def get_beams(self,refl,cond,opts='A'):
    #     # for i,F in enumerate(self.nFrames):
    #     rpl = self.rpl.loc[self.rpl.eval(cond)]

    ###########################################################################
    #### Show Integration
    ###########################################################################
    def show_Iavg(self,fz=np.log10,**kwargs):
        hkl = self.HKL_dyn.loc[self.HKL_dyn.I>1].copy()

        hkl['q0'] = np.round(hkl['rq']*100)/100
        qs0 = np.unique(hkl['q0'])
        Iavg0 = np.zeros(qs0.shape)
        for i,q0 in enumerate(qs0):
            Iavg0[i] = hkl.loc[hkl.q0==q0,'I'].mean()
        plts = [[qs0,fz(Iavg0),'b','']]

        hkl['q1'] = np.round(hkl['rq']*10)/10
        qs1 = np.unique(hkl['q1'])
        Iavg1 = np.zeros(qs1.shape)
        for i,q1 in enumerate(qs1):
            Iavg1[i] = hkl.loc[hkl.q1==q1,'I'].mean()

        plts+= [[qs1,fz(Iavg1),'r','']]
        dsp.stddisp(plts,labs=['$q(A^{-1})$','$I_{avg}$'],lw=2,#name=name+'_Iavg.svg',
            **kwargs)

    def show_Ihkl(self,dyn:Iterable=None,**kwargs):
        '''Show integrated intensity and average integrated intensity as
         function of resolution

         parameters
         -----------
         dyn
            show dynamical refinement reflections
        '''

        hkl = self.HKL_dyn.copy()
        rq = hkl['rq']
        I  = hkl['I']
        hkl['q']  = np.round(rq*10)/10
        hkl['order'] = np.zeros(rq.shape)
        qs = np.unique(hkl['q'])
        for i,q in enumerate(qs):
            beams = hkl.q==q
            hklq  = hkl.loc[beams][['h','k','l']].values
            order = np.argsort(np.sum(hklq,axis=1))
            hkl.loc[beams,'order'] = order

        cond = 'I>1'
        rq,order,I = hkl.loc[hkl.eval(cond),['rq','order','I']].values.T
        if type(dyn)==type(None):
            cmap,c = 'jet',np.log10(I)
        else:
            hkl['color']=0
            # h_dyn = [h for h in dyn if h in hkl.index]
            hkl.loc[dyn,'color']=1
            c = hkl.loc[hkl.eval(cond),'color'].values
            cmap='RdYlGn'
        # print(rq.shape,I.shape,c.shape,hkl.shape)
        scat = [rq,order,10*I**(0.2),c]

        xylims = [0.1,2,-1,order.max()+1]
        fig,ax = dsp.stddisp(scat=scat,labs=['$q(A^{-1})$','beam order'],xyTicks=[qs,[]],xylims=xylims,
            cs='S',sargs={'cmap':cmap},imOpt='c',opt='')

        # res = np.round(100/qs)/100
        res = np.array([0.5,0.6,0.7,0.8,0.9,1,1.5,2,3,5])
        q0,res = 1/res,np.array(res,dtype=str)
        dsp.addxAxis(ax,[],xLab=r'resolution $(A)$',c=(0.5,)*3,
            xTicks=q0,xTickLabs=res,xylims=xylims,axPos='V',#name=name+'_Ihkl.svg',
            **kwargs)

        fig.canvas.mpl_connect('button_press_event', self._on_click)

    def _on_click(self,event):
        from matplotlib.backend_bases import MouseButton
        if event.button is MouseButton.RIGHT:
            x,y = event.xdata,event.ydata
            idx = np.argmin(np.linalg.norm(self.hkl[['q','order']].values-[x,y],axis=1))
            hkl = self.hkl.iloc[idx]
            print(hkl[['rq','I']])

    def show_Iframes(self,frames,cm='hsv',refl=[],cond='',Imax=500,opts='AH',fz=abs,
        **kwargs):
        # frames = np.arange(1,self.nFrames+1)[iFs]
        refl = [str(tuple(h)) for h in refl]
        df = self.cor
        if 'A' in opts:
            df0 = df.loc[df.F<=frames[-1]]
            df0 = df0.loc[df0.eval(cond)]
            refl = list(pd.unique([str(tuple(h)) for h in df0[['h','k','l']].values]))
        if 'H' in opts:refl = [h for h in refl if h in self.hkl.index]
        # print(refl)

        nfs,nbs = np.array(frames).size,len(refl)
        I  = pd.DataFrame(np.nan*np.ones((nfs,nbs)),columns=[str(h) for h in refl])
        for i,F in enumerate(frames):
            # condF = '(F==%d) & ' %F + cond #;print(condF)
            dfF = df.loc[df.F==F]
            hkl = [str(tuple(h)) for h in dfF[['h','k','l']].values]#;print(hkl)
            # if not cond:
            idx = [i for i,refl0 in enumerate(hkl) if refl0 in refl]#;print(idx)
            hkl0 = np.array(hkl)[idx]                               #;print(hkl0)
            I.loc[i,hkl0] = dfF.iloc[idx]['I'].values

        I = I.loc[:,I.columns[I.max()<Imax]]
        refl = list(I.keys())
        Ib = fz(I.values.T)
        Imax = np.argmax(Ib,axis=1) #locate maxI

        cs,txts = dsp.getCs(cm,len(refl)),[]
        plts = [[frames,Ib[i],[cs[i],'-o'],'%s' %hkl] for i,hkl in enumerate(refl)]
        # if 't' in opts:txts = [[frames[idx],I[i,idx],'%s' %str(h),cs[i]] for i,(h,idx) in enumerate(zip(refl,Imax))]
        dsp.stddisp(plts,texts=txts,labs=['frame','$I$'],**kwargs)
        return I


    ###########################################################################
    #### display
    ###########################################################################
    def show_ewald_sphere(self,frame=None,Smax=0.01,Nmax=10,h3d=0,
        nts=100,nps=200,**kwargs):
        K = self.get_beam(frame)
        Kx,Ky,Kz = K

        (h,k,l),(qx,qy,qz) = self.get_lattice(Nmax)
        x,y,z = self.get_ewald(frame,nts,nps)
        df    = self.get_excitation_errors(frame,Smax=Smax,Nmax=Nmax)

        print('small excitation error beams',uvw)
        if frame:print('frame %d:' %frame)
        print(df[['h','k','l']])

        scat = ([qx,qy,qz,5,'b','o'],[0,0,0,'r','s'])
        scat+= ([df.qx,df.qy,df.qz,50,'g','s'],)
        surf = [x,y,z,{'alpha':0.4,'color':'r'}]
        plts = [[0,Kx],[0,Ky],[0,Kz],'r']
        # scat+=([x0,y0,z0])
        fig,ax = dsp.stddisp(plts,scat=scat,surfs=[surf],lw=3,
            labs=['x','y','z'],rc='3d',**kwargs)
        if h3d:h3d = h3D.handler_3d(fig,persp=False)

    def show_uvw(self):
        """show trajectory of orientation axis"""
        uvw = self.uvw
        ez =  np.cross(uvw[0],uvw[-1])
        tle = "orientation axis with frames"

        plts = [[[0,u[0]],[0,u[1]],[0,u[2]],'b'] for u in uvw ]
        plts +=[[ [0,ez[0]],[0,ez[1]],[0,ez[2]] , 'r' ]]
        dsp.stddisp(plots=plts,title=tle,labs=['x','y','z'],rc='3d',lw=4)

    def show_xyz(self,**kwargs):
        """show reflections in reciprocal space """
        x,y,z = self.XYZ
        tle = "reflections in reciprocal space"

        fig2,ax = dsp.stddisp(rc='3d',title=tle,scat=[x,y,z,2,'b'],labs=['x','y','z'],**kwargs)
        h3d2 = h3D.handler_3d(fig2,persp=False)

    def show_hkl(self,**kwargs):
        """Show intensities as hkl"""
        h,k,l,I = self.HKL[['h','k','l','I']].values.T
        fig2,ax = dsp.stddisp(rc='3d',scat=[h,k,l,I,'b'],labs=['h','k','l'],**kwargs)
        h3d2 = h3D.handler_3d(fig2,persp=False)

    def show_exp(self,frame=1,**kwargs):
        """show the experimental images"""
        return vw.Pets_Viewer(self,frame=frame,**kwargs)
    def show_sim(self,frame=1,**kwargs):
        """show the simulated images"""
        return vw.Pets_Viewer(self,frame=frame,sim=True,**kwargs)

    # def mp4_crystal_rotation(self,name,**kwargs):
    #     cif_file = self.cif_file
    #     uvw = [self.get_beam_dir(frame=f) for f in np.arange(self.nFrames)+1]
    #     for i,n in enumerate(uvw[:1]):
    #         frame = i+1
    #         name_x = 'figures/%s_x_%s.png' %(name,str(frame).zfill(3))
    #         name_y,name_z = name_x.replace('_x_','_y_'),name_x.replace('_x_','_z_')
    #         mut.show_cell(cif_file,name=name_x,title='%d' %frame,n=n,view=[0,0]  ,h3D=0,opt='sc',**kwargs)
    #         mut.show_cell(cif_file,name=name_y,title='%d' %frame,n=n,view=[90,0] ,h3D=0,opt='sc',**kwargs)
    #         mut.show_cell(cif_file,name=name_z,title='%d' %frame,n=n,view=[90,90],h3D=0,opt='sc',**kwargs)
    #
    #     cmd  = '/bin/bash -i -c "im2mp4 figures/%s_x_%%03d.png figures/x.mp4" && ' %name
    #     cmd += '/bin/bash -i -c "im2mp4 figures/%s_y_%%03d.png figures/y.mp4" && ' %name
    #     cmd += '/bin/bash -i -c "im2mp4 figures/%s_z_%%03d.png figures/z.mp4" '    %name
    #     p = Popen(cmd,shell=True,stdout=PIPE,stderr=PIPE);p.wait()
    #     o,e = p.communicate();
    #     print(o.decode())

    # def show_sim(self,frame=1,**kwargs):
    #     mut.Viewer(self.path+'/multislice',frame=frame,**kwargs)
    #
    # def show_frames(self,thick=1000,Nmax=5,Smax=0.02,e0=[1,0,0],rot=0,Imag=10,Itol=20,opts='PKqr',**sargs):
    #     kargs = dict(zip(
    #         ['Nmax','Smax','e0','rot','Itol','opts'],
    #         [Nmax,Smax,e0,rot,Itol,opts]))
    #     return mut.Frames_Viewer(self,thick,Imag,kargs,**sargs)
    #
    # def show_frame(self,frame=0,opts='Pqr',
    #     thick=1000,Nmax=5,Smax=0.02,e0=[1,0,0],rot=0,Imag=10,Itol=20,
    #     show_hkl=True,qopt=True,rings=True,sim_pattern=None,
    #     **kwargs):
    #     ''' Show a frame with information specified by opts
    #     - opts : E(exp), P(proc), S(sim), K(kin), h(hkl),q(rec A),r(rings), k(hkl_k)
    #     '''
    #     if isinstance(opts,str):exp,proc,sim,kin,show_hkl,qopt,rings,show_hkl_kin = [c in opts for c in 'EPSKhqrk']
    #     if kin:qopt=1
    #
    #     plts,scat,txts,labs,qmax,wm = [[0,0,'b+']],(),[],['px','py'],0,5
    #     # plts=[]
    #     if qopt :
    #         wm*=self.aper
    #         labs = ['$q_x$','$q_y$']
    #     if proc:
    #         rpl0  = self.rpl.loc[self.rpl.F==frame]
    #         I,Im = rpl0[['I','Im']].values.T
    #         if qopt :
    #             px,py = rpl0[['qx','qy']].values.T
    #             # px,py = rpl0[['rpx','rpy']].values.T
    #             # cx,cy = self.cen.iloc[frame-1][['px','py']].values.T
    #             # px,py = ((np.vstack([px,-py]).T-[cx,-cy])*self.aper).T
    #         else:
    #             px,py = rpl0[['rpx','rpy']].values.T
    #             cx,cy = self.cen.iloc[frame-1][['px','py']].values.T
    #             plts = [[cx,cy,'b+']]
    #         c = (0.5,)*3
    #         scat += ([px,py,I,c,'o'],)
    #         qmax = np.ceil(max(py.max(),px.max()))
    #
    #         if show_hkl:
    #             hkl1  = rpl0[['h','k','l']]
    #             h,k,l = hkl1.values.T
    #             hx,kx,lx = rpl0[['hx','kx','lx']].values.T
    #             sx = lambda x:['%d' %round(x),'%.1f' %x][abs(x-round(x))>0.06]
    #             txts += [[x+0*wm,y+wm,'(%s,%s,%s)' %(sx(h0),sx(k0),sx(l0)),(0.5,)*3] for x,y,h0,k0,l0 in zip(px,py,hx,kx,lx)]
    #             print(h,k,l)
    #
    #     if kin:
    #         px_k,py_k,I_k,hkl_k = self.get_kin(frame,thick,Nmax,Smax,e0,rot,Imag)
    #         scat += ([px_k,py_k,I_k,'g','o'],)
    #         if show_hkl_kin:
    #             h,k,l = hkl_k
    #             Itol = I_k.max()/Itol
    #             txts += [[x,y+wm,'(%s,%s,%s)' %(h0,k0,l0),'g'] for x,y,h0,k0,l0,I in zip(px_k,py_k,h,k,l,I_k) if I>Itol]
    #             print(h,k,l)
    #         qmax = np.ceil(max(qmax,px_k.max(),py_k.max()))
    #
    #     im,bgcol = None,'k'
    #     if sim:
    #         qx,qy,I=sim_pattern
    #         qmax = np.ceil(max(qmax,qx.max(),qy.max()))
    #         im,bgcol=[qx,-qy,I],None
    #         # print(bgcol)
    #
    #     if rings and qopt:
    #         t = np.linspace(0,np.pi*2,100)
    #         ct,st = np.cos(t),np.sin(t)
    #         plts+=[[i*ct,i*st,'m','',0.5] for i in np.arange(0.25,qmax,0.25)]
    #         plts+=[[i*ct,i*st,'m','',2] for i in np.arange(1,qmax)]
    #     # if rot:
    #     #     ct,st = np.cos(np.deg2rad(rot)),np.sin(np.deg2rad(rot))
    #     #     px,py = ct*px-st*py,st*px+ct*py
    #
    #
    #     if not 'fonts' in kwargs.keys():kwargs['fonts']={'text':15}
    #     if not 'xylims' in kwargs.keys():kwargs['xylims']=[-px.max(),px.max(),-py.max(),py.max()]
    #     dsp.stddisp(plts,ms=20,scat=scat,im=im,texts=txts,bgcol=bgcol,gridOn=0,
    #         labs=labs,sargs={'alpha':0.5},**kwargs)

    ###########################################################################
    #### private compares
    ###########################################################################
    def _rotate_pattern(self,frame,qxqy):
        xyz0  = np.vstack([qx,qy,np.zeros(qx.shape)])

        alpha,beta,omega = -self.cif.iloc[frame-1][['alpha','beta','omega']]

        R = np.identity(3)
        omega_r = np.deg2rad(omega)
        ct,st  = np.cos(omega_r),np.sin(omega_r)
        Rz = np.array([[ct,st,0],[-st,ct,0],[0,0,1]]) #get_crystal_rotation(u=[0,0,-1],alpha=self.omega)
        R = Rz.dot(R)
        alpha_r = np.deg2rad(alpha)
        ct,st   = np.cos(alpha_r),np.sin(alpha_r)
        Rx = np.array([[1,0,0],[0,ct,st],[0,-st,ct]])
        R  = Rx.dot(R)
        beta_r = np.deg2rad(beta)
        ct,st   = np.cos(beta_r),np.sin(beta_r)
        Ry = np.array([[ct,0,st],[0,1,0],[-st,0,ct]])
        R  = Ry.dot(R)

        x0,y0,z0  = R.dot(xyz0)

    def _qxyz_to_pxy(self,frame,qxyz):
        alpha,beta,gamma = self.cif.iloc[frame-1][['alpha','beta','omega']]
        alpha_r,beta_r,gamma_r = -np.deg2rad([alpha,beta,gamma])
        ctx,stx = np.cos(alpha_r),np.sin(alpha_r)
        cty,sty = np.cos(beta_r) ,np.sin(beta_r)
        ctz,stz = np.cos(gamma_r),np.sin(gamma_r)

        Rx = np.array([[1,0,0],[0,ctx,stx],[0,-stx,ctx]])
        Ry = np.array([[cty,0,sty],[0,1,0],[-sty,0,cty]])
        Rz = np.array([[ctz,stz,0],[-stz,ctz,0],[0,0,1]])
        R = Ry.dot(Rx.dot(Rz))
        pxy  = R.dot(qxyz)[:2,:]
        return pxy

    def _compare_xyz_pxpy(self,frame=32,opts='oa',view=[90,90],**kwargs):
        rpl0    = self.rpl.loc[self.rpl.F==frame]
        pxc,pyc = self.cen.iloc[frame-1][['px','py']].values.T
        omega,beta,alpha = self.omega,0,rpl0.alpha.iloc[0]


        px,py = rpl0[['rpx','rpy']].values.T
        px    =  self.aper*(px-pxc)
        py    = -self.aper*(py-pyc)
        xyz0  = np.vstack([px,py,np.zeros(px.shape)])

        R = np.identity(3)
        if 'o' in opts:
            omega_r = np.deg2rad(omega)
            ct,st  = np.cos(omega_r),np.sin(omega_r)
            Rz = np.array([[ct,st,0],[-st,ct,0],[0,0,1]]) #get_crystal_rotation(u=[0,0,-1],alpha=self.omega)
            R = Rz.dot(R)
        if 'a' in opts:
            alpha_r = np.deg2rad(alpha)
            ct,st   = np.cos(alpha_r),np.sin(alpha_r)
            Rx = np.array([[1,0,0],[0,ct,st],[0,-st,ct]])
            R  = Rx.dot(R)
        if 'b' in opts:
            beta_r = np.deg2rad(beta)
            ct,st   = np.cos(beta_r),np.sin(beta_r)
            Ry = np.array([[ct,0,st],[0,1,0],[-st,0,ct]])
            R  = Ry.dot(R)

        x0,y0,z0  = R.dot(xyz0)

        xyz   = rpl0[['x','y','z']].values.T
        x,y,z = xyz

        Im = np.array(rpl0.Im.values/10,dtype=int)
        if isinstance(view,str):
            if   view=='x':labs,scat = ['y','z'],([y,z,Im,'b','o'],[y0,z0,Im,'r','o'])
            elif view=='y':labs,scat = ['x','z'],([x,z,Im,'b','o'],[x0,z0,Im,'r','o'])
            elif view=='z':labs,scat = ['x','y'],([x,y,Im,'b','o'],[x0,y0,Im,'r','o'])
            dsp.stddisp(xylims=1.5,scat=scat,labs=labs,legElt={'pxpy':'bo','xyz':'ro'},**kwargs)
        else:
            scat = ([x,y,z,Im,'b','o'],[x0,y0,z0,Im,'r','o'])
            fig,ax  = dsp.stddisp(rc='3d',view=view,xylims=1.5,scat=scat,labs=['x','y','z'],**kwargs)
            h3d = h3D.handler_3d(fig,persp=False)

    def _compare_hkl(self,frame,eps=None,Smax=0.025,Nmax=5,v=0):
        uvw  = self.get_beam_dir(frame=frame)
        df   = self.get_excitation_errors(frame,Nmax=Nmax,Smax=Smax)
        hkl0 = df[['h','k','l']]

        hkl1 = self.rpl.loc[self.rpl.F==frame] #[['hx','kx','lx']].values
        if eps:
            idx  = (abs(hkl1.hx-hkl1.h)<eps) & (abs(hkl1.kx-hkl1.k)<eps) & (abs(hkl1.lx-hkl1.l)<eps)
            hkl1 = hkl1.loc[idx]
        hkl1 = hkl1[['h','k','l']]

        rrmv = [r for r in hkl1.values if not np.where(np.linalg.norm(r-hkl0.values,axis=1)==0)[0].size]
        rrmv = np.array(rrmv,dtype=int)
        print('%d reflections in diffraction patterns ' %hkl1.shape[0])
        print('%d reflections within Smax=%.2E of the Ewald sphere ' %(hkl0.shape[0],Smax))
        print('%d reflections in diffraction pattern not in ewald construction ' %(rrmv.shape[0]))
        if rrmv.shape[0]:print('\tdetails : ',rrmv)

        ridx = []
        for r in hkl1.values:
            idx = np.where(np.linalg.norm(r-hkl0.values,axis=1)==0)[0]
            if idx.size:
                ridx+=[idx[0]]
        if v:
            print('Reflections from diffraction patterns :');print(hkl1)
            print('Reflections from excitation errors of Ewald construction :');
            print(df.iloc[ridx][['h','k','l','Sw']])
        return df.iloc[ridx][['h','k','l','Sw']]

    def _check_orientation_matrix(self):
        lab = np.identity(3)
        cry = self.A.dot(self.lattice_vectors)
        # cr2 = self.A.dot(self.reciprocal_vectors).T).T
        plts  = [ [[0,ai[0]],[0,ai[1]],[0,ai[2]],c,'$%s$' %l]  for i,(ai,c,l) in enumerate(zip(lab,['r','g','b'],['x','y','z'])) ]
        # plts += [ [[0,ai[0]],[0,ai[1]],[0,ai[2]],[c,'--'],'$%s$' %l]  for i,(ai,c,l) in enumerate(zip(cry,['r','g','b'],['a','b','c'])) ]
        x0,y0,z0 = [0.5]*3
        plts += [ [[x0,ai[0]+x0],[y0,ai[1]+y0],[z0,ai[2]+z0],[c,'--'],'$%s^{*}$' %l]  for i,(ai,c,l) in enumerate(zip(cr2,['r','g','b'],['a','b','c'])) ]
        # dsp.stddisp(plts,rc='3d',view=[0,0],name='figures/glycine_orient.png',opt='sc')


def gauss2D(X, amp, x0, y0, sx,sy,noise):
    x,y = X
    g = noise + amp*np.exp(-((x-x0)/sx)**2 - ((y-y0)/sy)**2)
    return g
f_gauss2D = lambda X,amp,x0,y0,sx,sy,noise:gauss2D(X,amp, x0, y0, sx,sy,noise).ravel()




def make_pets(pts_file:str,
    aperpixel:float,alphas:Sequence[float]=None,
    deg:float=None,
    ref_cell:Sequence[float]=None,
    pxy:Optional[tuple] = (256,)*2,
    dmM:Optional[tuple] = (0.05,3),
    tag:Optional[str]=''):
    """creates a .pts file from info to process a simulated experiment with pets

    Parameters
    ----------
    pts_file
        full path to the pts file to write
    aperpixel
        aperture per pixel
    deg
        step angle between frames
    ref_cell
        cell info [ax,by,cz,alpha,beta,gamma] if known
    tag
        tag for the tiff images if there is one
    """
    center,pixel_size='AUTO',0.01
    if not deg:deg=alphas[1]-alphas[0]
    phi,omega= deg/2 ,0
    ax,by,cz,alpha,beta,gamma = ref_cell
    px,py = pxy
    dm,dM = dmM
    # ax,by,cz,alpha,beta,gamma = 8.1218, 9.977, 17.725, 90.0, 90.0, 90.0
    pts = '''lambda 0.025080
geometry continuous
omega  %d
phi  %.5f
virtualframes   7 5 1

aperpixel    %.6f
noiseparameters      2.5000      1.0000
saturationlimit   20000

center    %s %s
centermode friedelpairs 0

beamstop no


dstarmax  3
dstarmaxps  %s
dstarmin  %s
i/sigma    7.00    5.00
reflectionsize  7

referencecell     %.5f    %.5f     %.5f    %.5f   %.5f    %.5f 1

#List of images
#columns: file name,alpha,beta,delta omega,frame scale,calibration correction(px/rec.angstrom),ellipt.distortion correction(amplitude),ellipt.distortion correction(phase), use for calculations(0/1)
imagelist
'''%(omega,phi,aperpixel,px,py,dM,dm,  ax,by,cz,alpha,beta,gamma)
    out = os.path.dirname(pts_file)
    tif_files = np.sort(glob.glob(out+'/tiff/%s*.tiff' %tag))
    # print(tif_files)
    if type(alphas)==type(None):
        alphas = np.arange(tif_files.size)*deg

    for i,tif_file in enumerate(tif_files):
        tif_file = os.path.basename(tif_file)
        pts+='%s %.4f 0.0 0.0 1.0 0 0 0  1\n' %('tiff\\'+tif_file,alphas[i])
    pts+='endimagelist\n'
    # pts_file = out+name+'.pts'
    with open(pts_file,'w') as f:
        f.write(pts)
        print(colors.green+'file saved : '+colors.yellow+pts_file+colors.black)



import pandas as pd
import numpy as np


fv = lambda X,X0 : V0*np.exp(-5*2*np.linalg.norm(X-X0,axis=1)**2)
class ED_base():
    '''Base class for simluating ED patterns (z is the beam axis)
    **\ncrystal related**:\n
    - data : x,z,Zatom
    - Nx,Nz : number of unit cells
    - ppopt:V(potential)Q(Psi_q)
    '''
    def __init__(self,data,Nx=1,Nz=1,
        ppopt=''):
        self.data=data
        self.Nx=Nx
        self.Nz=Nz

        if 'V' in ppopt:self.potential_show()
    ##################################################################
    ### display
    def potential_show(self,**kwargs):
        x,z,Za = self.data
        C = [Cs[Z] for Z in self.Za]
        Nx0,Nx1 = int(self.Nx/2),int(np.ceil(self.Nx/2))
        dsp.stddisp(scat=[self.z,self.x,C],
            ms=20,labs=[r'$z(\AA)$',r'$x(\AA)$'],
            xyTicks=[self.bz,self.ax],xylims=[0,self.Nz*self.bz,-Nx0*self.ax,Nx1*self.ax],
            **kwargs)

    def F_show(self,**kwargs):
        Za = np.unique(self.Za)
        t = np.linspace(-self.tmax,self.tmax,100)
        tdeg = t*180/np.pi
        plts = [[tdeg,fj(t,Z),'b','$%d$' %Z] for Z in Za]
        dsp.stddisp(plts,labs=[r'$\theta (deg)$','$f_j$'],**kwargs)

    def getAngle(self):return (self.x0s/self.z0)*180/np.pi
    def getQ(self):return (self.x0s/self.z0)/self.lam
    def getI(self):return self.I



import setuptools

with open("README.md", "r") as fh:
   long_description = fh.read()

setuptools.setup(
    name="ccp4ED",
    version="1.5",
    author="Tarik Ronan Drevon",
    author_email="tarik.drevon@stfc.ac.uk",
    description="Electron diffraction utilities",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://pypi.org/project/ccp4ED",
    project_urls={
        'Documentation': 'https://debloch.readthedocs.io/en/latest/',
        'Source':'https://github.com/ccp4/electron-diffraction',
    },

    # packages=['multislice','scattering','wallpp'],
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={'multislice/data':['splines.npy']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License ",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['numpy','scipy','matplotlib','colorama','pandas',
    'crystals','TarikDrevonUtils','easygui','tifffile','pickle5','bs4',
    'cbf','mrcfile','gemmi',
    ],
)



import pickle,easygui,os
import pandas as pd
import numpy as np
# import matplotlib.pyplot as plt
from utils import glob_colors as colors

class KeyBinder:
    def __init__(self,name=''):
        if not name:
            name=os.path.join(os.path.dirname(__file__),'data','viewer')
        d = pd.read_pickle(name+'_dict.pkl')
        self.df_keys = pd.read_pickle(name+'_df.pkl')

        self.df_generic = self.df_keys.loc[d['generic']]
        self.df_modes   = self.df_keys.loc[d['modes']]
        self.df_orient  = self.df_keys.loc[d['orient']+d['dorient']]
        self.df_frames  = self.df_keys.loc[d['frame'] +d['dframe']]
        self.df_thicks  = self.df_keys.loc[d['thick'] +d['dthick']]
        self.df_bright  = self.df_keys.loc[d['brightness']+d['dbrightness']+d['rbrightness']]
        self.df_rot     = self.df_keys.loc[d['rot']]
        self.df_vals    = self.df_keys.loc[d['vals']]
        # self.df_bloch   = self.df_keys.loc[d['bloch']]
        self.df_pets    = self.df_keys.loc[d['dsp']+d['opt']]
        self.df_multi   = self.df_keys.loc[d['multi']]
        # self.df_pets['alias'].iloc[:4] = Pnum_keys
        self.dict_k = self.df_keys.key.to_dict()
        self.dict_a = self.df_keys.alias.to_dict()

        for type_k,list_k in d.items():
            self.__dict__[type_k+'_keys'] = list(self.df_keys.loc[list_k].key.values)

        # print(self.dsp_keys)
        self.pets_keys = self.dsp_keys+self.opt_keys
        self.opt_chars = dict(zip(self.opt_keys,'hkrl'))
        self.dsp_chars = dict(zip(self.dsp_keys,'PMBKVSL'))

        #### Keys that trigger the show
        self.show_keys = self.modes_keys.copy()
        self.show_keys += self.pets_keys
        self.show_keys += self.rot_keys+self.thick_keys+self.brightness_keys+self.rbrightness_keys
        self.show_keys +=[self.dict_k[k] for k in ['settings','hold','solve_bloch']]

    def is_key(self,key):return self.df_keys.loc[(self.df_keys.key==key) | (self.df_keys.alias==key)]
    def is_cmd(self,cmd):return self.df_keys.iloc[[cmd in k for k in  self.df_keys.index]]

    def __call__(self,event):
        self.shortcut_call(event.key)

    def shortcut_call(self,key):
        combs = ['control','shift','alt',
         'ctrl+shift','ctrl+alt','ctrl+alt+shift',
         'shift+control','shift+super','shift+ctrl+super','shift+alt+control',
         'alt+control','alt+shift']
        if not key in combs :
            dfk = self.is_key(key)
            if any(dfk.index):print(colors.green+'found key'+colors.black);print(dfk)
            else:print(colors.red+'no key binding for %s' %key+colors.black)


    ###################################
    ##### Help
    ###################################
    def show_help_cli(self):
        print(colors.green+'Shortcuts : '+colors.black)
        print(colors.blue+'Generic :                    '+colors.black);print(self.df_generic)
        print(colors.blue+'Modes :                      '+colors.black);print(self.df_modes)
        print(colors.blue+'Frames (Frames mode) :       '+colors.black);print(self.df_frames)
        print(colors.blue+'Orientation (Rotate mode) :  '+colors.black);print(self.df_orient)
        print(colors.blue+'Thickness :                  '+colors.black);print(self.df_thicks)
        print(colors.blue+'Brightness :                 '+colors.black);print(self.df_bright)
        print(colors.blue+'Show values :                '+colors.black);print(self.df_vals)
        print(colors.blue+'Display:                     '+colors.black);print(self.df_pets)
        # print(colors.blue+'Display (Rotate mode): '+colors.black);print(self.df_bloch)

    def show_help_gui(self):
        # msg = str(self.df_keys)
        msg =''
        msg+= '\t\t\t SHORTCUTS : \n'
        msg+= 'Generic :     '+'\n'
        msg+=str(self.df_generic)+'\n'
        msg+= 'Modes :       '+'\n'
        msg+=str(self.df_modes)+'\n'
        msg+= 'Frames (Frames mode) :      '+'\n'
        msg+=str(self.df_frames)+'\n'
        msg+= 'Orientation (Rotate mode) : '+'\n'
        msg+=str(self.df_orient)+'\n'
        msg+= 'Thickness :   '+'\n'
        msg+=str(self.df_thicks)+'\n'
        msg+= 'Brightness :  '+'\n'
        msg+=str(self.df_bright)+'\n'
        msg+= 'Show values : '+'\n'
        msg+=str(self.df_vals)+'\n'
        # msg+= 'Display (Rotate mode): '+'\n'
        # msg+=str(self.df_bloch)+'\n'
        msg+= 'Display: '+'\n'
        msg+=str(self.df_pets)+'\n'
        easygui.msgbox(msg, title='help')

    def show_help_simple(self):
        print(self.df_keys)

################################################################################
#### keymapping
################################################################################
def save_keys(name):
    dict_k,dict_a = {},{}

    generic  = ['new_fig'   ,'settings' ,'help_cli','help_gui','help_simple' ]
    gen_keys = ['ctrl+enter','enter'    ,'ctrl+2'  ,'ctrl+1'  ,'ctrl+ '      ]
    generic += ['save_config','save_config_as','open_config','reload_config',]
    gen_keys+= ['ctrl+s'     ,'ctrl+S'        ,'ctrl+o'     ,'backspace'    ,]
    generic += ['save_fig','save_bloch','solve_bloch']
    gen_keys+= ['s'       ,'ctrl+alt+S','ctrl+B'     ]
    generic += ['hold','find_shortcut','find_command']
    gen_keys+= [' '   ,'ctrl+P'       ,'ctrl+p'      ]
    dict_k.update(dict(zip(generic,gen_keys)))

    multi      = ['open_multi','solve_multi','set_multi','run_multi','show_multi','show_tags' ,'log_info']
    multi_keys = ['ctrl+O'    ,'m'          ,'ctrl+m'   ,'ctrl+M'   ,'ctrl+alt+m','ctrl+alt+M','ctrl+l'  ]
    dict_k.update(dict(zip(multi,multi_keys)))

    #modes
    modes = ['frames_mode','rotate_mode']
    # modes_keys = ['ctrl+alt+'+c for c in '12']
    # modes_keys = ['ctrl+shift+'+c for c in '12']
    modes_keys = ['!','"']
    dict_k.update(dict(zip(modes,modes_keys)))

    #orientation
    orient = ['elev_up','elev_down','azim_up','azim_down']
    orient_keys  = ['up','down','right','left']
    dict_k.update(dict(zip(orient,orient_keys)))
    dorient = ['d'+s for s in orient]
    dorient_keys = ['ctrl+'+s for s in orient_keys]
    dict_k.update(dict(zip(dorient,dorient_keys)))

    #rotate in frame
    rots  = ['rot_up','rot_down','rotS_up','rotS_down']
    rot_keys  = ['r','R','ctrl+r','ctrl+R']
    dict_k.update(dict(zip(rots,rot_keys)))
    # drots = ['d'+s for s in rots]
    # dorient_keys = ['ctrl+'+s for s in orient_keys]
    # dict_k.update(dict(zip(drots,dorient_keys)))

    #frames
    frames = ['frame_'+s for s in ['up','down']] #'right','down','left']]
    dframes = ['inc_'+s for s in frames]
    frame_keys    = ['up','down']
    dframe_keys   = ['ctrl+'+s for s in frame_keys]
    frame_keys_a  = ['right','left']
    # dframe_keys_a = ['ctrl+'+s for s in frame_keys_a]
    dict_k.update(dict(zip(frames,frame_keys)))
    dict_k.update(dict(zip(dframes,dframe_keys)))
    dict_a.update(dict(zip(frames,frame_keys_a)))
    frame_keys = frame_keys.copy()+frame_keys_a

    #thickness
    thicks  = ['thick_up','thick_down']
    dthicks = ['d'+s for s in thicks]
    thick_keys  = ['shift+ctrl+up','shift+ctrl+down']
    dthick_keys = ['ctrl+t','ctrl+T']
    dict_k.update(dict(zip(thicks,thick_keys)))
    dict_k.update(dict(zip(dthicks,dthick_keys)))

    #brightness
    bright  = ['brightness '+s for s in ['down','up']]
    dbright = ['d'+s for s in bright]
    rbright = [s+'brightness reset' for s in ['','d']]
    brightness_keys  = ['pagedown','pageup']
    dbrightness_keys = ['ctrl+'+k for k in brightness_keys]
    rbrightness_keys = ['shift+ctrl+'+k for k in brightness_keys]
    dict_k.update(dict(zip(bright,brightness_keys)))
    dict_k.update(dict(zip(dbright,dbrightness_keys)))
    dict_k.update(dict(zip(rbright,rbrightness_keys)))

    #vals
    vals = ['hkl','I','z','u','v','beams_vs_thickness','Ig','Vg','Sw']
    vals = ['show_'+s for s in vals]
    vals_keys = ['ctrl+'+c for c in 'hizuvbIVE']
    dict_k.update(dict(zip(vals,vals_keys)))

    # #display Bloch
    # Fs    = ['L','Sw','Vg','S','I','Ig']#,'E']
    # fopt  = ['m','L' ,'l' ,'m','m','m' ]#,'m']
    # bloch = ['Lattice','Excitation error','Potential','Scattering','Intensities','Kinematic']
    # bloch = ['show '+s for s in bloch]
    # F_keys    = [''+c for c in 'LGVSIK']
    # Fnum_keys = [''+c for c in '123456']
    # df_F = pd.DataFrame.from_dict(dict(zip(['names','key','alias','F','fopt'],[bloch,F_keys,Fnum_keys,Fs,fopt])))
    # df_F = df_F.set_index('names')

    #display Frames
    dict_k['show_Exp']='0'
    petsD = ['Processed','Multislice','Bloch','Kinematic','Potential','Excitation error','Lattice']
    petsD = ['show_'+s for s in petsD]
    dsp_keys   = ['P','M','B','K','V','S','L']
    dsp_keys_a = [''+c for c in '1234567']
    dsp_chars = dict(zip(dsp_keys,'PMBKVSL'))
    dict_k.update(dict(zip(petsD,dsp_keys)))
    dict_a.update(dict(zip(petsD,dsp_keys_a)))
    #options
    petsO = ['hkl_exp','hkl_kin','rings','legend']
    petsO = ['show_'+s for s in petsO]
    opt_keys = ['j','k','g','l']
    opt_chars = dict(zip(opt_keys,'hkrl'))
    dict_k.update(dict(zip(petsO,opt_keys)))


    #### Data frame for the help command
    df_keys = pd.DataFrame.from_dict(dict_k,orient='index',columns=['key'])
    # df_keys = df_keys.append(df_F[['key','alias']])

    df_keys['alias'] = ['']*df_keys.shape[0]
    for k,v in dict_a.items():df_keys.loc[k,'alias'] = v
    df_keys['mode']  = ['any']*df_keys.shape[0]
    df_keys.loc[orient+dorient,'mode'] = 'rotate'
    df_keys.loc[frames+dframes,'mode'] = 'frames'

    d = dict(zip(['generic','modes','frame','dframe','orient','dorient','rot','thick','dthick','brightness','dbrightness','rbrightness','vals','dsp'  ,'opt'  ,'multi'],
                 [ generic , modes , frames, dframes, orient , dorient , rots, thicks, dthicks, bright     , dbright     , rbright     , vals , petsD , petsO , multi ]
                 ))

    df_keys.to_pickle(name+'_df.pkl')
    with open(name+'_dict.pkl','wb') as f:pickle.dump(d,f,pickle.HIGHEST_PROTOCOL)
    print(colors.yellow+name+colors.green+' saved'+colors.black)
    # return df_keys,d

if __name__=='__main__':
    save_keys(name='data/viewer')



# import importlib as imp
import os,pickle5
import numpy as np, pandas as pd
import matplotlib.pyplot as plt
from matplotlib import patches
from utils import displayStandards as dsp                       #;imp.reload(dsp)
from utils import glob_colors as colors
from scipy.interpolate import interp2d,RegularGridInterpolator
from scipy.spatial import cKDTree
from scattering import structure_factor as sf
# from matplotlib import collections as coll
# from scipy.spatial import Delaunay
# import matplotlib.tri as mtri
from . import lattice
from . import config as cg
# from . import nodesTri as nt

class Wallpaper :
    '''Class for creating wallpapers'''
    def __init__(self,pp_type,a=1,b=1,alpha=90,pattern=[[0,0,0]],
        pOpt='',nh=3,nk=3,ndeg=100,interp=True,fract=False,
        name='',path='') :
        ''' initialize wallpaper
        pp_type   : plane group type (str)
        a,b,alpha : lattice parameters (double)
        pattern   : Nx3 array : x y f1 [f2 ...] coordinates in unit cells
        fract     : if pattern is given in fractional coordinates
        pOpt      : plot options (see plot_unit_cells)
        '''
        self.pp_type  = pp_type
        self.name     = name if name else pp_type #cg.df_wallpp.loc[pp_type]['long_name']
        self.path     = path
        self.df = cg.df_wallpp.loc[pp_type]
        self._init_lattice(a,b,alpha)
        self._init_unit_cell()
        self._init_pattern(pattern,fract)

        self.pattern  = cg.generate(self.pp_type,self.lat_vec,self.pattern0)
        self.pattern_fract = self.pattern.copy()
        self.pattern_fract[:,:2]=self.pattern_fract[:,:2].dot(self.lat1)
        if interp:self._interp_pattern()
        if pOpt:self.plot_wallpp(pOpt,nh,nk)

    def _init_lattice(self,a,b,alpha) :
        self.a,self.b,self.alpha = a,b,alpha
        self.lat_type = cg.df_wallpp.loc[self.pp_type]['lat_type']
        self.lat      = lattice.Lattice2D(self.lat_type,a,b,alpha)
        self.params   = self.lat.params
        self.lat_vec            = self.lat.lat_vec
        self.lattice_vectors    = self.lat_vec
        self.lat1               = np.linalg.inv(self.lat_vec)
        self.rec_vec            = self.lat.rec_vec
        self.reciprocal_vectors = self.rec_vec
        self.area       = self.lat.area
        a1,a2           = self.lat_vec
        self.unit_cell  = np.array([0*a1,a1,a1+a2,a2])

    def _init_unit_cell(self) :
        verts    = self.df.asym_cell
        a,b,angle = self.params
        self.asym_cell = verts.dot(self.lat_vec)

    def _init_pattern(self,pattern,fract):
        self.pattern0_fract = np.array(pattern).copy()
        if not fract:
            self.pattern0_fract[:,:2] = self.pattern0_fract[:,:2].dot(self.lat1)
        self.pattern0 = self.pattern0_fract.copy()
        self.pattern0[:,:2] = self.pattern0[:,:2].dot(self.lat_vec)

    def _interp_pattern(self):
        pattern = self.pattern0.copy()
        pattern[:,:2] = pattern[:,:2].dot(self.lat1)                    #fractional coordinates
        pattern1 = cg.generate_symmetries(self.pp_type,pattern=pattern) #generate
        x1,y1 = pattern1[:,:2].T
        z1 = pattern1[:,2:]
        # dsp.stddisp(scat=[x1,y1,z1[:,1]],pOpt='Xet');plt.show()

        print(colors.blue+'...nearest neighbor ...'+colors.black)
        ndeg = 100
        x0 = np.linspace(0,1,ndeg)
        x,y = np.meshgrid(x0,x0)
        xy  = np.array([x.flatten(),y.flatten()]).T
        idx = np.array([np.linalg.norm(X-pattern1[:,:2],axis=1).argmin() for X in xy])
        idx = np.reshape(idx,(ndeg,ndeg))
        # print(idx[0],idx[-1],idx.shape,idx.max())

        print(colors.blue+'...2D interpolation ...'+colors.black)
        # print(z1.shape,idx.max())
        # data = np.reshape(z1[idx,0],x.shape)
        # print(data.shape,x.shape)
        self.f = [RegularGridInterpolator((x0,x0), np.reshape(z[idx],x.shape),method='nearest') for z in z1.T]

        # self.f = [interp2d(x0,x0,z[idx]) for z in z1.T]
        # self.f = [interp2d(x1,y1,z) for z in z1.T]

    ######################################################################
    ##### display
    ######################################################################
    def show_ref(self,ndeg=100):
        print(colors.blue+'...plotting ref ...'+colors.black)
        x = np.linspace(0,1,ndeg)
        im = np.stack([f(x,x) for f in self.f],axis=2)
        im = np.array(im*255,dtype=int)#;print(im.max())
        plt.imshow(im);plt.show()
        # dsp.stddisp(im=[im])

    def show_xygrid(self,x,y,**kwargs):
        x0,y0 = x.copy(),y.copy()
        x,y = np.meshgrid(x0,y0)
        xy = x.flatten(),y.flatten()
        xy = np.array([x-np.floor(x),y-np.floor(y)]) .T
        im = np.stack([np.reshape(f(xy),x.shape) for f in self.f],axis=2)

        im = np.array(im*255,dtype=int)#;print(im.max())
        plt.imshow(im);plt.show()

    def plot(self,opts='aug',nh=3,nk=3,ms=5,pOpt='Xe',ax=None,**kwargs):
        '''plot wallpaper
        - opts : a(asym_cell),u(unit_cell),g(lattice grid)
        - nh,nk : multiplicity
        '''
        if not ax:fig,ax = dsp.stddisp(opt='')
        hk,R     = self.lat.get_lattice(nh,nk)
        R        = np.hstack([R,np.zeros((R.shape[0],self.pattern.shape[1]-2))])
        pattern1 = np.vstack([self.pattern +r for r in R])
        x,y = pattern1[:,:2].T
        z   = pattern1[:,2:]
        dsp.stddisp(ax=ax,scat=[x,y,z],ms=ms,pOpt=pOpt,opt='')

        pp = []
        if 'u' in opts:
            pp += [patches.Polygon(self.unit_cell,alpha=0.25,color='b')]
        if 'a' in opts:
            pp += [patches.Polygon(self.asym_cell,alpha=0.25,color='r')]
        if 'g' in opts:
            self.lat.plot_lattice(nh,nk,ax=ax,gridcolor='k',pOpt=pOpt,opt='')

        dsp.stddisp(ax=ax,patches=pp,pOpt=pOpt,**kwargs)

    def show_structure_factor(self,sf_args={},fz=abs,**kwargs):
        pattern = self.pattern.copy()
        pattern[:,:2] = pattern[:,:2].dot(self.lat1)
        (h,k),Fhk = sf.structure_factor2D(pattern,self.rec_vec*(2*np.pi),
            **sf_args)
        # hl = np.vstack([h.flatten(),k.flatten()]).T
        # qx,qy = hl.dot(self.rec_vec).T
        # qx = np.reshape(qx,Fhk.shape),np.reshape(qy,Fhk.shape)
        b1,b2 = self.rec_vec
        qx,qy = h*b1[0]+k*b2[0],h*b1[1]+k*b2[1]
        # print(qx.shape,Fhk.shape)
        dsp.stddisp(scat=[qx,qy,50,fz(Fhk)],labs=['$q_x(A)$','$q_y(A)$'],**kwargs)

    ######################################################################
    ##### misc
    ######################################################################
    def save(self,file=None):
        file = os.path.join(self.path,self.name+'_wallpp.pkl')
        with open(file,'wb') as f:pickle5.dump(self,f,pickle5.HIGHEST_PROTOCOL)
        print(colors.green+'file saved : ' + colors.yellow+ file+colors.black)

######################################################################
##### misc
######################################################################
def load(path='',name='',file=None):
    if name:file = os.path.join(path,name+'_wallpp.pkl')
    with open(file,'rb') as f : obj = pickle5.load(f)
    return obj
